# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import json
from copy import deepcopy
from os.path import isfile

DEBUG = False

# Global experiment placeholder. Should be set by the latest call of Experiment.init()
experiment = None

# Global Optimizer
optimizer = None


# TODO: Allow the configuration file path to be set by a environment variable?
CONFIG_PATH = ".comet_ml.json"

DEFAULT_CONFIG = {"uploaded_extensions": [".py", ".txt", ".json"]}

DEFAULT_UPLOAD_SIZE_LIMIT = 100 * 1024 * 1024  # 100 MebiBytes


def read_config_file(config, config_path):
    """ read and parse the JSON comet_ml configuration file """
    if isfile(config_path):
        with open(config_path) as config_file:
            raw_config = json.load(config_file)

            config.update(raw_config)

    return config


def get_config():
    """ return a copy of the default configuration
    """
    return deepcopy(DEFAULT_CONFIG)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

"""
Author: Boris Feld

This module contains the functions dedicated to logging the environment
information

"""

import getpass
import logging
import os
import platform
import socket
import sys

import netifaces
from comet_ml import connection
from six.moves.urllib.parse import urlparse

LOGGER = logging.getLogger(__name__)


def get_pid():
    return os.getpid()


def get_hostname():
    return socket.gethostname()


def get_os():
    return platform.platform(aliased=True)


def get_os_type():
    return platform.system()


def get_python_version_verbose():
    return sys.version


def get_python_version():
    return platform.python_version()


def get_user():
    return getpass.getuser()


def get_network_interfaces_ips():
    try:
        ips = []
        for interface in netifaces.interfaces():
            for link in netifaces.ifaddresses(interface).get(netifaces.AF_INET, []):
                ips.append(link["addr"])
        return ips

    except Exception:
        LOGGER.warning("Failed to log all interfaces ips", exc_info=True)
        return None


def get_ip():
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        parsed = urlparse(connection.server_address)
        host = parsed.hostname
        port = parsed.port
        if port is None:
            port = {"http": 80, "https": 443}.get(parsed.scheme, 0)
        s.connect((host, port))
        addr = s.getsockname()[0]
        s.close()
        return addr

    except socket.error:
        LOGGER.warning("Failed to log ip", exc_info=True)
        return None


def get_command():
    return sys.argv


def get_env_details():
    return {
        "pid": get_pid(),
        "hostname": get_hostname(),
        "os": get_os(),
        "os_type": get_os_type(),
        "python_version_verbose": get_python_version_verbose(),
        "python_version": get_python_version(),
        "user": get_user(),
        "network_interfaces_ips": get_network_interfaces_ips(),
        "ip": get_ip(),
        "command": get_command(),
    }



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

"""comet-ml"""
from __future__ import print_function

import atexit
import inspect
import os
import sys
import requests
import time
import os.path
import traceback
from contextlib import contextmanager
from copy import copy
import logging

import six
from pkg_resources import DistributionNotFound, get_distribution

from .file_uploader import (
    upload_file_process, upload_file_contents_process, upload_repo_start_process
)
from .comet import (
    Message,
    Streamer,
    config,
    get_cmd_args_dict,
    generate_guid,
    ENV_OPTIMIZATION_ID,
    is_valid_guid,
    save_matplotlib_figure,
)
from .config import get_config, DEFAULT_UPLOAD_SIZE_LIMIT
from .connection import (
    RestServerConnection,
    server_address,
    visualization_upload_url,
    OptimizerConnection,
    WebSocketConnection,
    INITIAL_BEAT_DURATION,
)
from .console import get_std_logger
from .exceptions import (
    NotParametrizedException,
    OptimizationMultipleParams,
    PCSParsingError,
    ValidationError,
    NoMoreSuggestionsAvailable,
    AuthenticationError,
    InvalidAPIKey,
)
from .keras_logger import patch as keras_patch
from .monkey_patching import CometModuleFinder
from .optimization import Suggestion, parse_pcs
from ._reporting import EXPERIMENT_CREATED, EXPERIMENT_CREATION_FAILED
from .sklearn_logger import patch as sklearn_patch
from .tensorboard_logger import patch as tb_patch
from .utils import is_list_like
from ._logging import (
    setup,
    INTERNET_CONNECTION_ERROR,
    IPYTHON_NOTEBOOK_WARNING,
    METRIC_ARRAY_WARNING,
    EXPERIMENT_OPTIMIZER_API_KEY_MISMTACH_WARNING,
    PARSING_ERR_MSG,
    NOTEBOOK_MISSING_ID,
    LOG_IMAGE_OS_ERROR,
    LOG_IMAGE_TOO_BIG,
    LOG_FIGURE_TOO_BIG,
    EXPERIMENT_LIVE,
    INVALID_API_KEY,
)
from ._notebook import (
    _jupyter_server_extension_paths,
    _jupyter_nbextension_paths,
    load_jupyter_server_extension,
    get_notebook_id,
    in_notebook_environment,
)
from .env_logging import get_env_details

try:
    __version__ = get_distribution("comet_ml").version
except DistributionNotFound:
    __version__ = "Please install comet with `pip install comet_ml`"

__author__ = "Gideon<Gideon@comet.ml>"
__all__ = ["Experiment"]

LOGGER = logging.getLogger(__name__)

# Activate the monkey patching
MODULE_FINDER = CometModuleFinder()
keras_patch(MODULE_FINDER)
sklearn_patch(MODULE_FINDER)
tb_patch(MODULE_FINDER)
MODULE_FINDER.start()

# Configure the logging
setup(
    os.environ.get("COMET_LOGGING_FILE"),
    os.environ.get("COMET_LOGGING_FILE_LEVEL", "INFO"),
)


def start():
    """
    If you are not using an Experiment in your first loaded Python file, you
    must import `comet_ml` and call `comet_ml.start` before any other imports
    to ensure that comet.ml is initialized correctly.
    """
    pass


class Experiment(object):
    """
    Experiment is a unit of measurable research that defines a single run with some data/parameters/code/results.

    Creating an Experiment object in your code will report a new experiment to your Comet.ml project. Your Experiment
    will automatically track and collect many things and will also allow you to manually report anything.

    You can create multiple objects in one script (such as when looping over multiple hyper parameters).

    """

    def __init__(
        self,
        api_key=None,
        project_name=None,
        team_name=None,
        log_code=True,
        log_graph=True,
        auto_param_logging=True,
        auto_metric_logging=True,
        parse_args=True,
        auto_output_logging="native",
        disabled=False,
        log_env_details=True,
    ):
        """
        Creates a new experiment on the Comet.ml frontend.
        Args:
            api_key: Your API key obtained from comet.ml
            project_name: Optional. Send your experiment to a specific project. Otherwise will be sent to `Uncategorized Experiments`.
                             If project name does not already exists Comet.ml will create a new project.
            team_name: Optional. Attach an experiment to a project that belongs to this team.
            log_code: Default(True) - allows you to enable/disable code logging
            log_graph: Default(True) - allows you to enable/disable automatic computation graph logging.
            auto_param_logging: Default(True) - allows you to enable/disable hyper parameters logging
            auto_metric_logging: Default(True) - allows you to enable/disable metrics logging
            parse_args: Default(True) - allows you to enable/disable automatic parsing of CLI arguments
            auto_output_logging: Default("native") - allows you to select
                which output logging mode to use. The default is `"native"`
                which will log all output even when it originated from a C
                native library. You can also pass `"simple"` which will work
                only for output made by Python code. If you want to disable
                automatic output logging, you can pass `None`.
            disabled: Default(False) - allows you to disable all network
                communication with the Comet.ml backend. It is useful when you
                just needs to works on your machine-learning scripts and need
                to relaunch them several times at a time.
            log_env_details: Default(True) - log various environment
                informations in order to identify where the script is running
        """
        self.project_name = project_name
        self.team_name = team_name
        if api_key is None:
            self.api_key = os.getenv("COMET_API_KEY", None)
        else:
            self.api_key = api_key

        if self.api_key is None:
            raise ValueError(
                "Comet.ml requires an API key. Please provide as the "
                "first argument to Experiment(api_key) or as an environment"
                " variable named COMET_API_KEY "
            )

        self.params = {}
        self.metrics = {}
        self.others = {}

        # Base config
        self.config = get_config()

        self.log_code = log_code
        self.log_graph = log_graph
        if in_notebook_environment():
            self.log_code = False
            LOGGER.debug("Notebook env detected, disabling logging code")
            notebook_id = get_notebook_id()

            if notebook_id is None:
                LOGGER.warning(NOTEBOOK_MISSING_ID)

        self.auto_param_logging = auto_param_logging
        self.auto_metric_logging = auto_metric_logging
        self.parse_args = parse_args
        self.auto_output_logging = auto_output_logging
        self.log_env_details = log_env_details

        # Generate a unique identifier for this experiment.
        self.id = self._get_experiment_key()
        self.alive = False
        self.is_github = False
        self.focus_link = None
        self.upload_limit = DEFAULT_UPLOAD_SIZE_LIMIT

        self.streamer = None
        self.logger = None
        self.run_id = None
        self.project_id = None

        # If set to True, wrappers should only run the original code
        self.disabled_monkey_patching = False

        # Experiment state
        self.context = None
        self.curr_step = None

        self.figure_counter = 0

        self.optimization_id = self._get_optimization_id_for_api_key(api_key)

        self.connection = RestServerConnection(
            self.api_key, self.id, self.optimization_id
        )

        # Cleanup old experiment before replace it
        if config.experiment is not None and config.experiment is not self:
            config.experiment._on_end(wait=False)

        config.experiment = self

        if disabled is not True:
            self._start()

            if self.alive is True:
                self.connection.report(event_name=EXPERIMENT_CREATED)

                LOGGER.info(EXPERIMENT_LIVE, self._get_experiment_url())

    def _get_experiment_key(self):
        return generate_guid()

    def _on_start(self):
        """ Called when the Experiment is started
        """
        self._mark_as_started()

    def _mark_as_started(self):
        try:
            self.connection.update_experiment_status(
                self.run_id, self.project_id, self.alive
            )
        except Exception:
            LOGGER.error("Failed to report experiment status", exc_info=True)

    def _mark_as_ended(self):
        if self.alive:
            try:
                self.connection.update_experiment_status(
                    self.run_id, self.project_id, False
                )
            except Exception:
                LOGGER.error("Failed to report experiment status", exc_info=True)

    def _on_end(self, wait=True):
        """ Called when the Experiment is replaced by another one or at the
        end of the script
        """
        self._mark_as_ended()

        if self.logger is not None:
            self.logger.clean()

        if self.streamer is not None:
            self.streamer.close()
            if wait is True:
                self.streamer.wait_for_finish()

    def _start(self):

        try:
            # This init the streamer and logger for the first time.
            # Would only be called once.
            if self.streamer is None and self.logger is None:
                full_ws_url = self._authenticate()

                # Authentication failed
                if full_ws_url is None:
                    return

                self._initialize_streamer_logger(full_ws_url)

            atexit.register(self._on_end)

            if self.logger:
                self.logger.set_experiment(self)
            self.alive = True

        except Exception as e:
            tb = traceback.format_exc()
            LOGGER.error("Run will not be logged", exc_info=True)
            self.connection.report(event_name=EXPERIMENT_CREATION_FAILED, err_msg=tb)
            return None

        try:
            if in_notebook_environment():
                self.set_notebook_name()
            else:
                self.filename = self._get_filename()
                self.set_filename(self.filename)
        except Exception as e:
            LOGGER.error("Failed to set run file name", exc_info=True)

        try:
            self.set_pip_packages()
        except Exception as e:
            LOGGER.error("Failed to set run pip packages", exc_info=True)

        try:
            if self.parse_args:
                self.set_cmd_args()
        except Exception as e:
            LOGGER.error("Failed to set run cmd args", exc_info=True)

        try:
            if self.log_code:
                self.set_code(self._get_source_code())
        except Exception as e:
            LOGGER.error("Failed to set run source code", exc_info=True)

        try:
            if self.log_env_details:
                self._log_env_details()
        except Exception as e:
            LOGGER.error("Failed to log environment details", exc_info=True)

        try:
            if self.log_code and self.is_github:
                self._upload_repository()
        except Exception as e:
            LOGGER.error("Failed to create git patch", exc_info=True)

        self._on_start()

    def _authenticate(self):
        """
        Do the handshake with the Backend to authenticate the api key and get
        various parameters and settings
        """
        # Get an id for this run
        try:
            self.run_id, full_ws_url, self.project_id, self.is_github, self.focus_link, self.upload_limit = self.connection.get_run_id(
                self.project_name, self.team_name
            )

            return full_ws_url

        except InvalidAPIKey as e:
            LOGGER.error(INVALID_API_KEY, e.api_key, exc_info=True)
            return

        except ValueError as e:
            tb = traceback.format_exc()
            LOGGER.error(INTERNET_CONNECTION_ERROR, exc_info=True)
            self.connection.report(event_name=EXPERIMENT_CREATION_FAILED, err_msg=tb)
            return

    def _initialize_streamer_logger(self, full_ws_url):
        """
        Initialize the streamer and logger with the websocket url received
        during the handshake.
        """
        # Initiate the streamer
        ws_connection = WebSocketConnection(full_ws_url, self.connection)
        ws_connection.start()
        ws_connection.wait_for_connection()
        self.streamer = Streamer(ws_connection, INITIAL_BEAT_DURATION, self.connection)

        if in_notebook_environment():
            # Don't hijack sys.std* in notebook environment
            self.logger = None
            LOGGER.warning(IPYTHON_NOTEBOOK_WARNING)
        else:
            # Override default sys.stdout and feed to streamer.
            self.logger = get_std_logger(self.auto_output_logging, self.streamer)
        # Start streamer thread.
        self.streamer.start()

    def _get_experiment_url(self):
        if self.focus_link:
            return self.focus_link + self.id

        return ""

    def _create_message(self):
        """
        Utility wrapper around the Message() constructor
        Returns: Message() object.

        """
        return Message(
            self.api_key,
            self.id,
            self.run_id,
            self.project_id,
            context=self.context,
            optimization_id=self.optimization_id,
            notebook_id=get_notebook_id(),
        )

    def get_metric(self, name):
        return self.metrics[name]

    def get_parameter(self, name):
        return self.params[name]

    def get_other(self, name):
        return self.others[name]

    def get_key(self):
        """
        Returns the experiment key, useful for using with the ExistingExperiment class
        Returns: Experiment Key (String)
        """
        return self.id


    def log_other(self, key, value):
        """
        Reports key,value to the `Other` tab on Comet.ml. Useful for reporting datasets attributes,
        datasets path, unique identifiers etc.

        See [`log_parameter`](#experimentlog_parameter)

        Args:
            key: Any type of key (str,int,float..)
            value: Any type of value (str,int,float..)

        Returns: None

        """
        if self.alive:
            message = self._create_message()
            message.set_log_other(key, value)
            self.streamer.put_messge_in_q(message)

        self.others[key] = value

    def log_html(self, html):
        """
        Reports any HTML blob to the `HTML` tab on Comet.ml. Useful for creating your own rich reports.
        The HTML will be rendered as an Iframe. Inline CSS/JS supported.
        Args:
            html: Any html string. for example:
            ```
            experiment.log_html('<a href="www.comet.ml"> I love Comet.ml </a>')
            ```

        Returns: None

        """
        if self.alive:
            message = self._create_message()
            message.set_html(html)
            self.streamer.put_messge_in_q(message)

    def set_step(self, step):
        """
        Sets the current step in the training process. In Deep Learning each step is after feeding a single batch
         into the network. This is used to generate correct plots on comet.ml. You can also pass the step directly when reporting [log_metric](#experimentlog_metric), and [log_parameter](#experimentlog_parameter).

        Args:
            step: Integer value

        Returns: None

        """

        if step is not None:
            self.curr_step = step

    def log_epoch_end(self, epoch_cnt, step=None):
        """
        Logs that the  epoch finished. required for progress bars.

        Args:
            epoch_cnt: integer

        Returns: None

        """
        self.set_step(step)

        if self.alive:
            message = self._create_message()
            message.set_param("curr_epoch", epoch_cnt, step=self.curr_step)
            self.streamer.put_messge_in_q(message)

    def log_metric(self, name, value, step=None):
        """
        Logs a general metric (i.e accuracy, f1).

        e.g.
        ```
        y_pred_train = model.predict(X_train)
        acc = compute_accuracy(y_pred_train, y_train)
        experiment.log_metric("accuracy", acc)
        ```

        See also [`log_multiple_metrics`](#experimentlog_multiple_metrics)


        Args:
            name: String - name of your metric
            value: Float/Integer/Boolean/String
            step: Optional. Used as the X axis when plotting on comet.ml

        Returns: None

        """
        LOGGER.debug("Log metric: %s %r %r", name, value, step)

        self.set_step(step)

        if self.alive:
            message = self._create_message()

            if is_list_like(value):
                # Try to get the first value of the Array
                try:
                    if len(value) != 1:
                        raise TypeError()

                    if not isinstance(
                        value[0], (six.integer_types, float, six.string_types, bool)
                    ):
                        raise TypeError()

                    value = value[0]

                except (TypeError):
                    LOGGER.warning(METRIC_ARRAY_WARNING, value)
                    value = str(value)

            message.set_metric(name, value, self.curr_step)
            self.streamer.put_messge_in_q(message)

        # save state.
        self.metrics[name] = value

    def log_parameter(self, name, value, step=None):
        """
        Logs a single hyperparameter. For additional values that are not hyper parameters it's encouraged to use [log_other](#experimentlog_other).

        See also [`log_multiple_params`](#experimentlog_multiple_params).


        Args:
            name: String - name of your parameter
            value: Float/Integer/Boolean/String/List
            step: Optional. Used as the X axis when plotting on comet.ml

        Returns: None

        """
        self.set_step(step)

        if self.alive:
            message = self._create_message()

            # Check if we have a list-like object or a string
            if is_list_like(value):
                message.set_params(name, value, self.curr_step)
            else:
                message.set_param(name, value, self.curr_step)

            self.streamer.put_messge_in_q(message)

        self.params[name] = value

    def log_figure(self, figure_name=None, figure=None):
        """
        Logs the global Pyplot figure or the passed one and upload its svg
        version to the backend.

        Args:
            figure_name: Optional. String - name of the figure
            figure: Optional. The figure you want to log. If not set, the
                    global pyplot figure will be logged and uploaded
        """
        if self.alive is False:
            return

        try:
            filename = save_matplotlib_figure(figure)
        except Exception as e:
            LOGGER.warning("Failing to save the matplotlib figure")
            # An error occured
            return

        # Check the filesize
        file_size = os.path.getsize(filename)
        if file_size > self.upload_limit:
            LOGGER.error(LOG_FIGURE_TOO_BIG, file_size, self.upload_limit)
            return

        # Pass additional url params
        figure_number = self.figure_counter
        self.figure_counter += 1
        url_params = {
            "step": self.curr_step,
            "figCounter": figure_number,
            "context": self.context,
            "runId": self.run_id,
        }

        if figure_name is not None:
            url_params["figName"] = figure_name

        upload_file_process(
            self.project_id,
            self.id,
            filename,
            visualization_upload_url(),
            self.api_key,
            url_params,
        )

    def log_image(self, file_path, file_name=None):
        """
        Logs the image located at given file_path. Images are displayed on the Graphics tab on Comet.ml

        Args:
            file_path: String - the file path of the image you wants to log
            file_name: String - Optional. A custom file name to be displayed
            on the dashboard.
        """
        if self.alive is False:
            return

        # Check the file size before reading it
        try:
            file_size = os.path.getsize(file_path)
            file_basename = os.path.basename(file_path)
            if file_size > self.upload_limit:
                LOGGER.error(LOG_IMAGE_TOO_BIG, file_path, file_size, self.upload_limit)
                return

            # Read the file contents synchronously to protect against file
            # modifications that could happens when log_image call ends and
            # before a subprocess access the file
            with open(file_path, "rb") as image_file:
                file_contents = image_file.read()
        except OSError:
            LOGGER.error(LOG_IMAGE_OS_ERROR, file_path)
            return

        # Prepare parameters
        figure_number = self.figure_counter
        self.figure_counter += 1
        url_params = {
            "step": self.curr_step,
            "context": self.context,
            "runId": self.run_id,
            "figName": file_name,
            "figCounter": figure_number,
        }

        upload_file_contents_process(
            self.project_id,
            self.id,
            file_contents,
            file_basename,
            visualization_upload_url(),
            self.api_key,
            url_params,
        )

    def log_current_epoch(self, value):
        if self.alive:
            message = self._create_message()
            message.set_metric("curr_epoch", value)
            self.streamer.put_messge_in_q(message)

    def log_multiple_params(self, dic, prefix=None, step=None):
        """
        Logs a dictionary of multiple parameters.
        See also [log_parameter](#experimentlog_parameter).

        e.g:
        ```
        experiment = Experiment(api_key="MY_API_KEY")
        params = {
            "batch_size":64,
            "layer1":"LSTM(128)",
            "layer2":"LSTM(128)",
            "MAX_LEN":200
        }

        experiment.log_multiple_params(params)
        ```
        """
        self.set_step(step)

        if self.alive:
            for k in sorted(dic):
                if prefix is not None:
                    self.log_parameter(prefix + "_" + str(k), dic[k], self.curr_step)
                else:
                    self.log_parameter(k, dic[k], self.curr_step)

    def log_multiple_metrics(self, dic, prefix=None, step=None):
        """
        Logs a key,value dictionary of metrics.
        See also [`log_metric`](#experimentlog_metric)
        """
        self.set_step(step)

        if self.alive:
            for k in sorted(dic):
                if prefix is not None:
                    self.log_metric(prefix + "_" + str(k), dic[k], self.curr_step)
                else:
                    self.log_metric(k, dic[k], step)

    def log_dataset_hash(self, data):
        try:
            import hashlib

            data_hash = hashlib.md5(str(data).encode("utf-8")).hexdigest()
            self.log_parameter("dataset_hash", data_hash[:12])
        except:
            LOGGER.warning("Failed to create dataset hash")

    def set_code(self, code):
        """
        Sets the current experiment script's code. Should be called once per experiment.
        Args:
            code: String. Experiment source code.
        """
        if self.alive:
            message = self._create_message()
            message.set_code(code)
            self.streamer.put_messge_in_q(message)

    def set_model_graph(self, graph):
        """
        Sets the current experiment computation graph.
        Args:
            graph: String. Usually JSON represented graph.
        """
        if self.alive:

            LOGGER.debug("Set model graph called")

            if type(graph).__name__ == "Graph":  # Tensorflow Graph Definition
                from google.protobuf import json_format

                graph_def = graph.as_graph_def()
                graph = json_format.MessageToJson(graph_def)

            message = self._create_message()
            message.set_graph(graph)
            self.streamer.put_messge_in_q(message)

    def set_filename(self, fname):
        """
        Sets the current experiment filename.
        Args:
            fname: String. script's filename.
        """
        if self.alive:
            message = self._create_message()
            message.set_filename(fname)
            self.streamer.put_messge_in_q(message)

    def set_name(self, name):
        """
        Set a name for the experiment. Useful for filtering and searching on Comet.ml.
        Will shown by default under the `Other` tab.
        Args:
            name: String. A name for the experiment.
        """
        self.log_other("Name", name)

    def set_notebook_name(self):
        self.set_filename("Notebook")

    def set_pip_packages(self):
        """
        Reads the installed pip packages using pip's CLI and reports them to server as a message.
        Returns: None

        """
        if self.alive:
            try:
                import pkg_resources

                installed_packages = [d for d in pkg_resources.working_set]
                installed_packages_list = sorted(
                    ["%s==%s" % (i.key, i.version) for i in installed_packages]
                )
                message = self._create_message()
                message.set_installed_packages(installed_packages_list)
                self.streamer.put_messge_in_q(message)
            except:
                LOGGER.warning("Failing to collect the installed pip packages")

    def set_cmd_args(self):
        if self.alive:
            args = get_cmd_args_dict()
            LOGGER.debug("Command line arguments %r", args)
            if args is not None:
                for k, v in args.items():
                    self.log_parameter(k, v)

    def set_uploaded_extensions(self, extensions):
        """
        Override the default extensions that will be sent to the server.

        Args:
            extensions: list of extensions strings
        """
        self.config["uploaded_extensions"] = copy(extensions)

    # Context context-managers

    @contextmanager
    def train(self):
        """
        A context manager to mark the beginning and the end of the training
        phase. This allows you to provide a namespace for metrics/params.
        For example:

        ```
        experiment = Experiment(api_key="MY_API_KEY")
        with experiment.train():
            model.fit(x_train, y_train)
            accuracy = compute_accuracy(model.predict(x_train),y_train) # returns the train accuracy
            experiment.log_metric("accuracy",accuracy) # this will be logged as train accuracy based on the context.
        ```

        """
        # Save the old context and set the new one
        old_context = self.context
        self.context = "train"

        yield self

        # Restore the old one
        self.context = old_context

    @contextmanager
    def validate(self):
        """
        A context manager to mark the beginning and the end of the validating
        phase. This allows you to provide a namespace for metrics/params.
        For example:

        ```
        with experiment.validate():
            pred = model.predict(x_validation)
            val_acc = compute_accuracy(pred, y_validation)
            experiment.log_metric("accuracy", val_acc) # this will be logged as validation accuracy based on the context.
        ```


        """
        # Save the old context and set the new one
        old_context = self.context
        self.context = "validate"

        yield self

        # Restore the old one
        self.context = old_context

    @contextmanager
    def test(self):
        """
        A context manager to mark the beginning and the end of the testing phase. This allows you to provide a namespace for metrics/params.
        For example:

        ```
        with experiment.test():
            pred = model.predict(x_test)
            test_acc = compute_accuracy(pred, y_test)
            experiment.log_metric("accuracy", test_acc) # this will be logged as test accuracy based on the context.
        ```

        """
        # Save the old context and set the new one
        old_context = self.context
        self.context = "test"

        yield self

        # Restore the old one
        self.context = old_context

    def get_keras_callback(self):
        """
        Returns an instance of Comet.ml's Keras callback. This callback is already added to your Keras `model.fit()` callbacks list automatically, to report model training metrics to Comet.ml.


        e.g:
        ```
        experiment = Experiment(api_key="MY_API_KEY")
        comet_callback = experiment.get_keras_callback()

        Returns: Comet.ml Keras callback.

        """
        if self.alive:
            from comet_ml.frameworks import KerasCallback

            return KerasCallback(
                self,
                log_params=self.auto_param_logging,
                log_metrics=self.auto_metric_logging,
                log_graph=self.log_graph,
            )

        from comet_ml.frameworks import EmptyKerasCallback

        return EmptyKerasCallback()

    def disable_mp(self):
        """ Disabling the auto-collection of metrics and monkey-patching of
        the Machine Learning frameworks.
        """
        self.disabled_monkey_patching = True

    def _get_source_code(self):
        """
        Inspects the stack to detect calling script. Reads source code from disk and logs it.
        """

        for frame in inspect.stack(context=1):
            module = inspect.getmodule(frame[0])
            if "comet_ml" != module.__name__:
                filename = module.__file__.rstrip("cd")
                with open(filename) as f:
                    return f.read()

        LOGGER.warning("Failed to find source code module")

    def _get_filename(self):

        if sys.argv:
            pathname = os.path.dirname(sys.argv[0])
            abs_path = os.path.abspath(pathname)
            filename = os.path.basename(sys.argv[0])
            full_path = os.path.join(abs_path, filename)
            return full_path

        return None

    def _upload_repository(self):
        upload_repo_start_process(
            self.project_id,
            self.id,
            self.filename,
            server_address + "logger/repoRoot",
            server_address + "logger/uploadFiles",
            self.api_key,
            config=self.config,
        )

    def _log_env_details(self):
        if self.alive:
            message = self._create_message()
            message.set_env_details(get_env_details())
            self.streamer.put_messge_in_q(message)

    def _get_optimization_id_for_api_key(self, api_key):
        if config.optimizer is not None:
            if config.optimizer.api_key == api_key:
                return config.optimizer.id

            else:
                LOGGER.warning(EXPERIMENT_OPTIMIZER_API_KEY_MISMTACH_WARNING)
        return None


class ExistingExperiment(Experiment):
    """
    Existing Experiment allows you to report information to an experiment that already exists on comet.ml
    and is not currently running. This is usfeul in cases where your training and testing happens on different scripts.

    For example:

    train.py:
    ```
    exp = Experiment(api_key="my-key")
    score = train_model()
    exp.log_metric("train accuarcy", score)
    ```

    Now obtain the experiment key from comet.ml. If it's not visible on your experiment table
    you can click `Customize` and add it as a column.


    test.py:
    ```
    exp = ExistingExperiment(api_key="my-key", previous_experiment = "your experiment key from comet.ml")
    score = test_model()
    exp.log_metric("test accuarcy", score)
    ```

    """

    def __init__(self, api_key, previous_experiment):
        # Validate the previous experiment id
        if not is_valid_guid(previous_experiment):
            raise ValueError("Invalid experiment key: %s" % previous_experiment)

        self.previous_experiment = previous_experiment
        super(ExistingExperiment, self).__init__(api_key)

    def _get_experiment_key(self):
        return self.previous_experiment

    def _authenticate(self):
        """
        Do the handshake with the Backend to authenticate the api key and get
        various parameters and settings
        """
        # Get an id for this run
        try:
            self.run_id, full_ws_url, self.project_id, self.is_github, self.focus_link, self.upload_limit = self.connection.get_old_run_id(
                self.previous_experiment
            )

            return full_ws_url

        except InvalidAPIKey as e:
            LOGGER.error(INVALID_API_KEY, e.api_key, exc_info=True)
            return

        except ValueError as e:
            LOGGER.error(INTERNET_CONNECTION_ERROR, exc_info=True)
            return

    def _create_message(self):
        # Create a default context to distinguish between script runs
        if self.context is None:
            new_context = generate_guid()
            self.context = new_context

        return super(ExistingExperiment, self)._create_message()


class Optimizer(object):
    '''
    An Optimizer is the object that you can use to dynamically optimize your hyper-parameters on the cloud.

    You can use it this way:

    ```
    optimizer = Optimizer("API_KEY")

    # Declare your hyper-parameters in the PCS format
    params = """
    x integer [1, 10] [10]
    y real [1, 10] [1.0]
    """

    optimizer.set_params(params)

    # get_suggestion will raise when no new suggestion is available
    while True:
        # Get a suggestion
        suggestion = optimizer.get_suggestion()

        # Create a new experiment associated with the Optimizer
        experiment = Experiment("API_KEY")

        # Test the model
        score = fit(suggestion["x"])

        # Report the score back
        suggestion.report_score("accuracy",score)
    ```

    '''

    def __init__(self, api_key):
        """
        Creates an Optimizer that you can use to get hyper-parameter
        suggestions.
        Args:
            api_key: Your API key obtained from comet.ml
        """
        self.api_key = api_key
        self.id = generate_guid()
        self.headers = {"OPTIMIZATION-ID": self.id}
        self.types = {}
        self.params_set = False
        self.connection = OptimizerConnection(self.headers)

        self._authenticate()

        # Save the optimization id in global config so Experiment can use it.
        config.optimizer = self

    def _authenticate(self):
        # Authenticate to the hyper-parameter service
        try:
            self.connection.authenticate(self.api_key, self.id)
        except requests.exceptions.HTTPError as e:
            response = e.response
            # If their was a validation error, print it
            if response.status_code == 400:
                try:
                    data = response.json()
                    error = data.get("msg")
                    if error:
                        six.raise_from(AuthenticationError(error), None)
                except ValueError:
                    pass

            raise

    def set_params_file(self, file_path):
        """
        Declare your hyper-parameter using a file in the PCS format.
        Args:
            file_path: The PCS file path
        """
        with open(file_path, "r") as pcs_file:
            pcs_content = pcs_file.read()
        self._send_params(pcs_content)

    def set_params(self, pcs_content):
        """
        Declare your hyper-parameter using a string in the PCS format.
        Args:
            pcs_content: A string in the PCS format. Leading spaces are not
                         significant.
        """
        self._send_params(pcs_content)

    def _send_params(self, pcs_content):
        """ Send the pcs content to the hyper-parameter optimization API and
        try to parse it to extract type information
        """
        # Check that we can send params only once
        if self.params_set is True:
            raise OptimizationMultipleParams()

        # Parse the pcs_content
        try:
            self.types = parse_pcs(pcs_content)
        except PCSParsingError:
            # Fallback on no types and print warning
            LOGGER.warning(PARSING_ERR_MSG)
            self.types = {}

        # Send the requests
        try:
            self.connection.create(pcs_content)

            self.params_set = True
        except requests.exceptions.HTTPError as e:
            response = e.response
            # If their was a validation error, print it
            if response.status_code == 400:
                try:
                    data = response.json()
                    error = data.get("error")
                    if error:
                        six.raise_from(ValidationError(error), None)
                except ValueError:
                    pass

            raise

    def get_suggestion(self):
        """ Return a new [Suggestion](Suggestion/#Suggestion) object
        containing values for each (TODO: Check with conditions) declare
        hyper-parameter.
        """

        if self.params_set is False:
            raise NotParametrizedException()

        try:
            while True:
                response = self.connection.get_suggestion()

                data = response.json()
                suggestion = data.get("suggestion")
                if suggestion:
                    return Suggestion(suggestion, self, self.types)

                else:
                    if data.get("terminated"):
                        raise NoMoreSuggestionsAvailable()

                    time.sleep(1)
        except requests.exceptions.HTTPError as err:
            exc_info = sys.exc_info()
            if err.response.status_code != 404:
                raise

            try:
                data = err.response.json()
                if data.get("error") == "Missing instance":
                    err_msg = "No optimization session has been found, did you called set_params or set_params_file before calling get_suggestion?"
                    six.raise_from(Exception(err_msg), None)

            except ValueError:
                raise err

    def _report_score(self, run_id, score):
        # Maximize the score
        score = -1 * score

        self.connection.report_score(run_id, score)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import sys

import six


def is_iterable(value):
    try:
        iter(value)
        return True

    except TypeError:
        return False


def is_list_like(value):
    """ Check if the value is a list-like
    """
    if is_iterable(value) and not isinstance(value, six.string_types):
        return True

    else:
        return False



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

from __future__ import print_function

import json
import logging
import os
import threading
import time

import comet_ml
import requests
import websocket
from comet_ml import config
from comet_ml._logging import WS_ON_CLOSE_MSG, WS_ON_OPEN_MSG
from comet_ml._reporting import WS_ON_CLOSE, WS_ON_ERROR, WS_ON_OPEN
from comet_ml.config import DEFAULT_UPLOAD_SIZE_LIMIT
from comet_ml.exceptions import InvalidAPIKey
from comet_ml.json_encoder import NestedEncoder
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

optimization_address = os.environ.get(
    "COMET_OPTIMIZATION_OVERRIDE", "https://optimizer.comet.ml/"
)
TIMEOUT = 10

OPTIMIZER_SESSION = None
INITIAL_BEAT_DURATION = 10000  # 10 second

LOGGER = logging.getLogger(__name__)


def _comet_version():
    try:
        version_num = comet_ml.__version__
    except NameError:
        version_num = None

    return version_num


def get_retry_strategy():

    # The total backoff sleeping time is computed like that:
    # backoff = 2
    # total = 3
    # s = lambda b, i: b * (2 ** (i - 1))
    # sleep = sum(s(backoff, i) for i in range(1, total + 1))

    return Retry(
        total=3,
        backoff_factor=2,
        method_whitelist=False,
        status_forcelist=[400, 404, 500, 502, 503, 504],
    )  # Will wait up to 24s


def get_backend_address():
    return os.environ.get("COMET_URL_OVERRIDE", "https://www.comet.ml/clientlib/")


server_address = get_backend_address()


def get_backend_session():
    session = requests.Session()

    return session


def get_backend_session_with_retry(backend_address):
    session = requests.Session()
    session.mount(backend_address, HTTPAdapter(max_retries=get_retry_strategy()))

    return session


def get_optimizer_session():
    global OPTIMIZER_SESSION
    if OPTIMIZER_SESSION is None:
        OPTIMIZER_SESSION = requests.Session()

    return requests.Session()


def json_post(url, session, headers, body, timeout):
    response = session.post(
        url=url, data=json.dumps(body), headers=headers, timeout=timeout
    )

    response.raise_for_status()
    return response


class RestServerConnection(object):
    """
    A static class that handles the connection with the server.
    """

    def __init__(self, api_key, experiment_id, optimization_id, server_address=None):
        self.api_key = api_key
        self.experiment_id = experiment_id
        self.optimization_id = optimization_id

        # Set once get_run_id is called
        self.run_id = None
        self.project_id = None

        if server_address is None:
            self.server_address = get_backend_address()
        else:
            self.server_address = server_address

        self.session = get_backend_session()
        self.retry_session = get_backend_session_with_retry(self.server_address)

    def heartbeat(self):
        """ Inform the backend that we are still alive
        """
        LOGGER.debug("Doing an heartbeat")
        return self.update_experiment_status(self.run_id, self.project_id, True)

    def update_experiment_status(self, run_id, project_id, is_alive):
        endpoint_url = self.server_address + "status-report/update"
        headers = {"Content-Type": "application/json;charset=utf-8"}

        payload = {
            "apiKey": self.api_key,
            "runId": run_id,
            "experimentKey": self.experiment_id,
            "projectId": project_id,
            "optimizationId": self.optimization_id,
            "is_alive": is_alive,
            "local_timestamp": int(time.time() * 1000),
        }

        r = self.session.post(
            url=endpoint_url, data=json.dumps(payload), headers=headers, timeout=TIMEOUT
        )

        if r.status_code != 200:
            raise ValueError(r.content)

        data = r.json()
        beat_duration = data.get("is_alive_beat_duration_millis")

        if beat_duration is None:
            raise ValueError("Missing heart-beat duration")

        return beat_duration

    def get_run_id(self, project_name, team_name):
        """
        Gets a new run id from the server.
        :param api_key: user's API key
        :return: run_id - String
        """
        endpoint_url = self.server_address + "logger/add/run"
        headers = {"Content-Type": "application/json;charset=utf-8"}

        payload = {
            "apiKey": self.api_key,
            "local_timestamp": int(time.time() * 1000),
            "projectName": project_name,
            "teamName": team_name,
            "libVersion": _comet_version(),
        }

        LOGGER.debug("Get run id URL: %s", endpoint_url)
        r = self.retry_session.post(
            url=endpoint_url, data=json.dumps(payload), headers=headers, timeout=TIMEOUT
        )

        if r.status_code != 200:
            if r.status_code == 400:
                # Check if the api key was invalid
                data = r.json()  # Raise a ValueError if failing
                if data.get("sdk_error_code") == 90212:
                    raise InvalidAPIKey(self.api_key)

            raise ValueError(r.content)

        res_body = json.loads(r.content.decode("utf-8"))

        return self._parse_run_id_res_body(res_body)

    def get_old_run_id(self, previous_experiment):
        """
        Gets a run id from an existing experiment.
        :param api_key: user's API key
        :return: run_id - String
        """
        endpoint_url = self.server_address + "logger/get/run"
        headers = {"Content-Type": "application/json;charset=utf-8"}

        payload = {
            "apiKey": self.api_key,
            "local_timestamp": int(time.time() * 1000),
            "previousExperiment": previous_experiment,
            "libVersion": _comet_version(),
        }
        LOGGER.debug("Get old run id URL: %s", endpoint_url)
        r = self.retry_session.post(
            url=endpoint_url, data=json.dumps(payload), headers=headers, timeout=TIMEOUT
        )

        if r.status_code != 200:
            if r.status_code == 400:
                # Check if the api key was invalid
                data = r.json()  # Raise a ValueError if failing
                if data.get("sdk_error_code") == 90212:
                    raise InvalidAPIKey(self.api_key)

            raise ValueError(r.content)

        res_body = json.loads(r.content.decode("utf-8"))

        return self._parse_run_id_res_body(res_body)

    def _parse_run_id_res_body(self, res_body):
        run_id_server = res_body["runId"]
        ws_full_url = res_body["ws_full_url"]

        project_id = res_body.get("project_id", None)

        is_github = bool(res_body.get("githubEnabled", False))

        focus_link = res_body.get("focusUrl", None)

        upload_limit = res_body.get("upload_file_size_limit_in_mb", None)

        if not (isinstance(upload_limit, int) and upload_limit > 0):
            upload_limit = DEFAULT_UPLOAD_SIZE_LIMIT
        else:
            # The limit is given in Mb, convert it back in bytes
            upload_limit = upload_limit * 1024 * 1024

        res_msg = res_body.get("msg")
        if res_msg:
            LOGGER.info(res_msg)

        # Save run_id and project_id around
        self.run_id = run_id_server
        self.project_id = project_id

        return run_id_server, ws_full_url, project_id, is_github, focus_link, upload_limit

    def report(self, event_name=None, err_msg=None):

        try:
            if event_name is not None:
                endpoint_url = notify_url()
                headers = {"Content-Type": "application/json;charset=utf-8"}
                # Automatically add the sdk_ prefix to the event name
                real_event_name = "sdk_{}".format(event_name)

                payload = {
                    "event_name": real_event_name,
                    "api_key": self.api_key,
                    "run_id": self.run_id,
                    "experiment_key": self.experiment_id,
                    "project_id": self.project_id,
                    "err_msg": err_msg,
                    "timestamp": int(time.time() * 1000),
                }

                LOGGER.debug("Report notify URL: %s", endpoint_url)

                json_post(endpoint_url, self.session, headers, payload, TIMEOUT / 2)

        except Exception as e:
            LOGGER.debug("Error reporting", exc_info=True)
            pass


class Reporting(object):

    @staticmethod
    def report(
        event_name=None,
        api_key=None,
        run_id=None,
        experiment_key=None,
        project_id=None,
        err_msg=None,
        is_alive=None,
    ):

        try:
            if event_name is not None:
                endpoint_url = notify_url()
                headers = {"Content-Type": "application/json;charset=utf-8"}
                # Automatically add the sdk_ prefix to the event name
                real_event_name = "sdk_{}".format(event_name)

                payload = {
                    "event_name": real_event_name,
                    "api_key": api_key,
                    "run_id": run_id,
                    "experiment_key": experiment_key,
                    "project_id": project_id,
                    "err_msg": err_msg,
                    "timestamp": int(time.time() * 1000),
                }

                json_post(
                    endpoint_url, get_backend_session(), headers, payload, TIMEOUT / 2
                )

        except Exception:
            LOGGER.debug("Failing to report %s", event_name, exc_info=True)


def notebook_source_upload(content_hash, json_model, api_key, notebook_path):
    session = get_backend_session()

    payload = {"apiKey": api_key, "notebookPath": notebook_path, "code": json_model}

    endpoint_url = "%sjupyter-notebook/source/add?notebookId=%s" % (
        server_address, content_hash
    )
    headers = {"Content-Type": "application/json;charset=utf-8"}

    response = session.post(
        url=endpoint_url, data=json.dumps(payload), headers=headers, timeout=TIMEOUT
    )
    response.raise_for_status()

    return response


class WebSocketConnection(threading.Thread):
    """
    Handles the ongoing connection to the server via Web Sockets.
    """

    def __init__(self, ws_server_address, connection):
        threading.Thread.__init__(self)
        self.priority = 0.2
        self.daemon = True
        self.name = "WebSocketConnection(%s)" % (ws_server_address)
        self.closed = False

        if config.DEBUG:
            websocket.enableTrace(True)

        self.address = ws_server_address
        self.ws = self.connect_ws(self.address)

        self.connection = connection

    def is_connected(self):
        if self.ws.sock is not None:
            return self.ws.sock.connected

        return False

    def connect_ws(self, ws_server_address):
        ws = websocket.WebSocketApp(
            ws_server_address,
            on_message=self.on_message,
            on_error=self.on_error,
            on_close=self.on_close,
        )
        ws.on_open = self.on_open
        return ws

    def run(self):
        while self.closed is False:
            try:
                self._loop()
            except Exception as e:
                LOGGER.debug("Run forever error", exc_info=True)
                # Avoid hammering the backend
                time.sleep(0.5)
        LOGGER.debug("WebSocketConnection has ended")

    def _loop(self):
        self.ws.run_forever()
        LOGGER.debug("Run forever has ended")

    def send(self, messages):
        """ Encode the messages into JSON and send them on the websocket
        connection
        """
        LOGGER.debug("SENDING %d WS messages", len(messages))
        data = self._encode(messages)
        self._send(data)

    def close(self):
        LOGGER.debug("Closing %r", self)
        self.closed = True
        self.ws.close()

    def _encode(self, messages):
        """ Encode a list of messages into JSON
        """
        messages_arr = []
        for message in messages:
            payload = {}
            # make sure connection is actually alive
            if message.stdout is not None:
                payload["stdout"] = message
            else:
                payload["log_data"] = message

            messages_arr.append(payload)

        data = json.dumps(messages_arr, cls=NestedEncoder, allow_nan=False)
        LOGGER.debug("ENCODED DATA %r", data)
        return data

    def _send(self, data):
        if self.ws.sock:
            self.ws.send(data)
            LOGGER.debug("Sending data done")
            return

        else:
            LOGGER.debug("WS not ready for connection")
            self.wait_for_connection()

    def wait_for_connection(self, num_of_tries=10):
        """
        waits for the server connection
        Args:
            num_of_tries: number of times to try connecting before giving up

        Returns: True if succeeded to connect.

        """
        if not self.is_connected():
            counter = 0

            while not self.is_connected() and counter < num_of_tries:
                time.sleep(1)
                counter += 1

            if not self.is_connected():
                raise ValueError("Could not connect to server after multiple tries. ")

        return True

    def on_open(self, ws):
        LOGGER.debug(WS_ON_OPEN_MSG)

        self.connection.report(event_name=WS_ON_OPEN, err_msg=WS_ON_OPEN_MSG)

    def on_message(self, ws, message):
        if message != "got msg":
            LOGGER.debug("WS msg: %s", message)

    def on_error(self, ws, error):
        error_type_str = type(error).__name__
        ignores = [
            "WebSocketBadStatusException",
            "error",
            "WebSocketConnectionClosedException",
            "ConnectionRefusedError",
            "BrokenPipeError",
        ]

        self.connection.report(event_name=WS_ON_ERROR, err_msg=repr(error))

        # Ignore some errors for auto-reconnecting
        if error_type_str in ignores:
            LOGGER.debug("Ignore WS error: %r", error, exc_info=True)
            return

        LOGGER.debug("WS on error: %r", error, exc_info=True)

    def on_close(self, *args, **kwargs):
        LOGGER.debug(WS_ON_CLOSE_MSG)
        self.connection.report(event_name=WS_ON_CLOSE, err_msg=WS_ON_CLOSE_MSG)


def notify_url():
    return server_address + "notify/event"


def visualization_upload_url():
    """ Return the URL to upload visualizations
    """
    return server_address + "visualizations/upload"


class OptimizerConnection(object):

    def __init__(self, headers):
        self.session = get_optimizer_session()
        self.headers = headers

    def authenticate(self, api_key, optimization_id):
        data = {
            "optimizationId": optimization_id,
            "apiKey": api_key,
            "libVersion": _comet_version(),
        }

        response = self.session.post(
            optimization_address + "authenticate", json=data, headers=self.headers
        )
        response.raise_for_status()
        return response

    def create(self, pcs_content):
        files = {"file": pcs_content}
        response = self.session.post(
            optimization_address + "create", files=files, headers=self.headers
        )
        response.raise_for_status()
        return response

    def get_suggestion(self):
        response = self.session.get(
            optimization_address + "suggestion/", headers=self.headers
        )
        response.raise_for_status()
        return response

    def report_score(self, run_id, score):
        response = self.session.post(
            optimization_address + "value/%s" % run_id,
            json={"value": score},
            headers=self.headers,
        )
        response.raise_for_status()
        return response



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

"""

This module contains reporting constants for Comet

"""

EXPERIMENT_CREATED = "experiment_created"

EXPERIMENT_CREATION_FAILED = "experiment_creation_failed"

FILE_UPLOADED_FAILED = "file_uploaded_failed"

REPO_UPLOADED = "repo_uploaded"

REPO_UPLOADED_FAILED = "repo_uploaded_failed"

ON_EXIT_DIDNT_FINISH_UPLOAD_SDK = "on_exit_didnt_finish_upload_sdk"

WS_ON_OPEN = "ws_on_open"

WS_ON_ERROR = "ws_on_error"

WS_ON_CLOSE = "ws_on_close"



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

"""
Author: Gideon Mendels

This module contains the main components of comet.ml client side

"""
from __future__ import print_function

import json
import logging
import re
import sys
import tempfile
import threading
import time
import uuid
from math import isnan

from comet_ml import config
from comet_ml._logging import EXPERIMENT_LIVE
from comet_ml._reporting import ON_EXIT_DIDNT_FINISH_UPLOAD_SDK
from comet_ml.json_encoder import NestedEncoder
from six.moves.queue import Empty, Queue

DEBUG = False
ENV_OPTIMIZATION_ID = "COMETML_OPTIMIZATION_ID_{api_key}"
LOGGER = logging.getLogger(__name__)


class CloseMessage(object):
    """ A special messag indicating Streamer to ends and exit
    """
    pass


class Streamer(threading.Thread):
    """
    This class extends threading.Thread and provides a simple concurrent queue
    and an async service that pulls data from the queue and sends it to the server.
    """

    def __init__(self, ws_connection, beat_duration, connection):
        threading.Thread.__init__(self)
        self.daemon = True
        self.messages = Queue()
        self.counter = 0
        self.name = "Streamer(%r)" % (ws_connection)
        self.ws_connection = ws_connection
        self.closed = False
        self.beat_duration = beat_duration / 1000.  # We get milliseconds
        self.last_beat = time.time()
        self.connection = connection

        LOGGER.debug("Streamer instantiated with ws url %s", self.ws_connection)
        LOGGER.debug("Streamer instantiated with duration %s", self.beat_duration)

    def put_messge_in_q(self, message):
        """
        Puts a message in the queue
        :param message: Some kind of payload, type agnostic
        """
        if message is not None:
            LOGGER.debug("Putting 1 message in queue")
            self.messages.put(message)

    def close(self):
        """
        Puts a None in the queue which leads to closing it.
        """
        if self.closed is True:
            LOGGER.debug("Streamer tried to be closed more than once")
            return

        # Send a message to close
        self.put_messge_in_q(CloseMessage())

        self.closed = True

    def run(self):
        """
        Continuously pulls messages from the queue and sends them to the server.
        """
        self.ws_connection.wait_for_connection()

        while True:
            out = self._loop()

            # Exit the infinite loop
            if isinstance(out, CloseMessage):
                break

        LOGGER.debug("Streamer has finished")

        # Close the websocket connection
        self.ws_connection.close()

        return

    def _loop(self):
        """
        A single loop of running
        """
        try:
            if self.ws_connection is not None and self.ws_connection.is_connected():
                messages = self.getn(1)

                if messages and isinstance(messages[0], CloseMessage):
                    return messages[0]

                if messages is not None:
                    LOGGER.debug(
                        "Got %d messages, %d still in queue",
                        len(messages),
                        self.messages.qsize(),
                    )
                    try:
                        self.ws_connection.send(messages)
                    except Exception:
                        LOGGER.debug("WS sending error", exc_info=True)

                try:
                    self._check_heartbeat()
                except Exception:
                    LOGGER.debug("Heartbeat error", exc_info=True)

        except Exception:
            LOGGER.debug("Unknown streaming error", exc_info=True)

    def _check_heartbeat(self):
        """
        Check if we should send an heartbeat
        """
        next_beat = self.last_beat + self.beat_duration
        now = time.time()
        if next_beat < now:
            LOGGER.debug("Doing an hearbeat")
            new_beat_duration = self.connection.heartbeat()
            LOGGER.debug("Getting a new heartbeat duration %d", new_beat_duration)
            self.beat_duration = new_beat_duration / 1000.  # We get milliseconds
            self.last_beat = time.time()

    def getn(self, n):
        """
        Pops n messages from the queue.
        Args:
            n: Number of messages to pull from queue

        Returns: n messages

        """
        try:
            msg = self.messages.get(
                timeout=self.beat_duration
            )  # block until at least 1
        except Empty:
            LOGGER.debug("No message in queue, timeout")
            return None

        if isinstance(msg, CloseMessage):
            return [msg]

        self.counter += 1
        msg.set_offset(self.counter)
        result = [msg]
        try:
            while len(result) < n:
                another_msg = self.messages.get(
                    block=False
                )  # dont block if no more messages
                self.counter += 1
                another_msg.set_offset(self.counter)
                result.append(another_msg)
        except Exception:
            LOGGER.debug("Exception while getting more than 1 message", exc_info=True)
        return result

    def wait_for_finish(self):
        """ Blocks the experiment from exiting until all data was sent to server OR 30 seconds passed."""

        msg = "Uploading stats to Comet before program termination (may take several seconds)"
        LOGGER.info(msg)

        # Wait maximum 2 minutes
        self._wait_for_empty(30)

        if not self.messages.empty():
            msg = "Still uploading stats to Comet before program termination (may take several seconds)"
            LOGGER.info(msg)
            self._wait_for_empty(30)

            if not self.messages.empty():
                self._wait_for_empty(60, verbose=True)

        if not self.messages.empty():
            remaining = self.messages.qsize()
            LOGGER.error(
                "Comet failed to send all the data back (%s messages)", remaining
            )

        if config.experiment is not None:
            if config.experiment.alive:
                LOGGER.info(EXPERIMENT_LIVE, config.experiment._get_experiment_url())
            else:
                LOGGER.warning("Failed to log run in comet.ml")
        else:
            LOGGER.warning("Failed to log run in comet.ml")

    def _wait_for_empty(self, timeout, verbose=False):
        """ Wait up to TIMEOUT seconds for the messages queue to be empty
        """
        end_time = time.time() + timeout

        while not self.messages.empty() and time.time() < end_time:
            if verbose is True:
                LOGGER.info("%d messages remaining to upload", self.messages.qsize())
            time.sleep(5)

    def _report_upload_not_finished(self):
        """ Called if the queue is not empty during exit
        """
        remaining = self.messages.qsize()

        if remaining == 0:
            return

        err_msg = "Comet failed to send all the data back %d"

        # Display the message
        LOGGER.error(err_msg, remaining)

        self.connection.report(
            event_name=ON_EXIT_DIDNT_FINISH_UPLOAD_SDK, err_msg=(err_msg % remaining)
        )


INFINITY = float("inf")


def fix_special_floats(value, _inf=INFINITY, _neginf=-INFINITY):
    """ Fix out of bounds floats (like infinity and -infinity) and Not A
    Number.
    Returns either a fixed value that could be JSON encoded or the original
    value.
    """

    try:
        # Check if the value is Nan, equivalent of math.isnan
        if isnan(value):
            return "NaN"

        elif value == _inf:
            return "Infinity"

        elif value == _neginf:
            return "-Infinity"

    except TypeError:
        # Ndarray cannot be compared
        return value

    return value


class Message(object):
    """
    A bean used to send messages to the server over websockets.
    """

    def __init__(
        self,
        api_key,
        experiment_key,
        run_id,
        project_id,
        context=None,
        optimization_id=None,
        notebook_id=None,
    ):
        self.apiKey = api_key
        self.experimentKey = experiment_key
        self.runId = run_id
        self.projectId = project_id
        self.local_timestamp = int(time.time() * 1000)

        # The following attributes are optional
        self.metric = None
        self.param = None
        self.params = None
        self.graph = None
        self.code = None
        self.stdout = None
        self.stderr = None
        self.offset = None
        self.fileName = None
        self.env_details = None
        self.html = None
        self.installed_packages = None
        self.log_other = None

        self.context = context
        self.optimization_id = optimization_id
        self.notebook_id = notebook_id

    def set_log_other(self, key, value):
        self.log_other = {"key": key, "val": value}

    def set_installed_packages(self, val):
        self.installed_packages = val

    def set_offset(self, val):
        self.offset = val

    def set_metric(self, name, value, step=None):
        safe_value = fix_special_floats(value)
        self.metric = {"metricName": name, "metricValue": safe_value, "step": step}

    def set_html(self, value):
        self.html = value

    def set_param(self, name, value, step=None):
        safe_value = fix_special_floats(value)
        self.param = {"paramName": name, "paramValue": safe_value, "step": step}

    def set_params(self, name, values, step=None):
        safe_values = list(map(fix_special_floats, values))
        self.params = {"paramName": name, "paramValue": safe_values, "step": step}

    def set_graph(self, graph):
        self.graph = graph

    def set_code(self, code):
        self.code = code

    def set_stdout(self, line):
        self.stdout = line
        self.stderr = False

    def set_stderr(self, line):
        self.stdout = line
        self.stderr = True

    def set_env_details(self, details):
        self.env_details = details

    def set_filename(self, fname):
        self.fileName = fname

    def to_json(self):
        json_re = json.dumps(
            self.repr_json(), sort_keys=True, indent=4, cls=NestedEncoder
        )
        return json_re

    def repr_json(self):
        return self.__dict__

    def __repr__(self):
        filtered_dict = [(key, value) for key, value in self.__dict__.items() if value]
        string = ", ".join("%r=%r" % item for item in filtered_dict)
        return "Message(%s)" % string

    def __str__(self):
        return self.to_json()

    def __len__(self):
        return len(self.to_json())


def get_cmd_args_dict():
    if len(sys.argv) > 1:
        try:
            return parse_cmd_args(sys.argv[1:])

        except ValueError as e:
            LOGGER.warning(
                "Failed to parse argv values. Falling back to naive parsing."
            )
            return parse_cmd_args_naive(sys.argv[1:])


def parse_cmd_args_naive(to_parse):
    vals = {}
    if len(to_parse) > 1:
        for i, arg in enumerate(to_parse):
            vals["run_arg_%s" % i] = str(arg)

    return vals


def parse_cmd_args(argv_vals):
    """
    Parses the value of argv[1:] to a dictionary of param,value. Expects params name to start with a - or --
    and value to follow. If no value follows that param is considered to be a boolean param set to true.(e.g --test)
    Args:
        argv_vals: The sys.argv[] list without the first index (script name). Basically sys.argv[1:]

    Returns: Dictionary of param_names, param_values

    """

    def guess_type(s):
        import ast

        try:
            value = ast.literal_eval(s)
            return value

        except (ValueError, SyntaxError):
            return str(s)

    results = {}

    current_key = None
    for word in argv_vals:
        word = word.strip()
        prefix = 0

        if word[0] == "-":
            prefix = 1
            if len(word) > 1 and word[1] == "-":
                prefix = 2

            if current_key is not None:
                # if we found a new key but haven't found a value to the previous
                # key it must have been a boolean argument.
                results[current_key] = True

            current_key = word[prefix:]

        else:
            word = word.strip()
            if current_key is None:
                # we failed to parse the string. We think this is a value but we don't know what's the key.
                # fallback to naive parsing.
                raise ValueError("Failed to parse argv arguments")

            else:
                word = guess_type(word)
                results[current_key] = word
                current_key = None

    if current_key is not None:
        # last key was a boolean
        results[current_key] = True

    return results


def save_matplotlib_figure(figure=None):
    """ Try saving either the current global pyplot figure or the given one
    and return None in case of error.
    """
    # Get the right figure to upload
    if figure is None:
        import matplotlib.pyplot

        # Get current global figure
        figure = matplotlib.pyplot.gcf()

    # Check if the figure is empty or not
    axes = figure.gca()
    if axes.has_data() is False:
        # TODO DISPLAY BETTER ERROR MSG
        msg = "Refuse to upload empty figure, please call log_figure before calling show"
        LOGGER.warning(msg)
        raise TypeError(msg)

    # Save the file to a tempfile but don't delete it, the file uploader
    # process will take care of it
    tmpfile = tempfile.NamedTemporaryFile(suffix=".svg", delete=False)
    figure.savefig(tmpfile, format="svg")

    return tmpfile.name


def generate_guid():
    """ Generate a GUID
    """
    return str(uuid.uuid4()).replace("-", "")


def is_valid_guid(guid):
    """ Validate a GUID
    """
    # The compiled regex is cached
    guid_regex = re.compile(r"[0-9a-f]{32}")
    return guid_regex.match(guid)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************


import logging
import sys

from comet_ml import config

LOGGER = logging.getLogger(__name__)
DISABLED = False


TENSORFLOW_ALREADY_IMPORTED_MSG = (
    "Please import comet before importing any Tensorflow modules"
)

def extract_from_add_summary(file_writer, summary, global_step):
    from tensorflow.core.framework import summary_pb2

    extracted_values = {}

    if isinstance(summary, bytes):
        summ = summary_pb2.Summary()
        summ.ParseFromString(summary)
        summary = summ

    for value in summary.value:
        field = value.WhichOneof("value")

        if field == "simple_value":
            extracted_values[value.tag] = value.simple_value

    return extracted_values, global_step


def add_summary_logger(real_add_summary):

    def wrapper(*args, **kwargs):
        ret_val = real_add_summary(*args, **kwargs)
        try:
            disabled_monkey_patching = config.experiment.disabled_monkey_patching
            if (
                config.experiment
                and disabled_monkey_patching is False
                and DISABLED is False
            ):
                LOGGER.debug("TENSORBOARD LOGGER CALLED")
                params, step = extract_from_add_summary(*args, **kwargs)
                config.experiment.log_multiple_metrics(params, step=step)
        except Exception as e:
            LOGGER.error(
                "Failed to extract parameters from add_summary()", exc_info=True
            )
        finally:
            return ret_val

    return wrapper


ADD_SUMMARY = [
    ("tensorflow.python.summary.writer.writer", "SummaryToEventTransformer.add_summary")
]


def patch(module_finder):
    if "tensorflow" in sys.modules or "tensorboard" in sys.modules:
        LOGGER.warning(TENSORFLOW_ALREADY_IMPORTED_MSG)

    # Register the fit methods
    for module, object_name in ADD_SUMMARY:
        module_finder.register(module, object_name, add_summary_logger)


if "tensorboard" in sys.modules:
    raise SyntaxError("Please import comet before importing any tensorboard modules")



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

"""
Author: Boris Feld

This module contains comet generated Exceptions

"""


class OptimizationMultipleParams(Exception):

    def __str__(self):
        return "You cannot call multiple time set_params or set_params file"


class NotParametrizedException(Exception):

    def __str__(self):
        return "Please call set_params or set_params_file before calling get_suggestion"


class ValidationError(Exception):
    pass


class AuthenticationError(Exception):
    pass


class NoMoreSuggestionsAvailable(Exception):

    def __str__(self):
        return "No more suggestions available!"


class PCSParsingError(Exception):
    """ Exception raised when parsing a PCS file
    """

    def __init__(self, faulty_line):
        self.faulty_line = faulty_line


class PCSCastingError(Exception):

    def __init__(self, value, pcs_type):
        self.value = value
        self.pcs_type = pcs_type


class OptimizationMissingExperiment(Exception):

    def __str__(self):
        return "You must create an experiment instance to perform optimization"


class InvalidAPIKey(Exception):

    def __init__(self, api_key):
        super(Exception, self).__init__()
        self.api_key = api_key



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************
#
"""
Author: Gideon Mendels

This module contains the components for console interaction like the std
wrapping

"""
import logging
import os
import sys

from comet_ml._logging import NATIVE_STD_WRAPPER_NOT_AVAILABLE, UNKOWN_STD_WRAPPER_SPEC

LOGGER = logging.getLogger(__name__)


class FakeStd(object):
    """ A fake Std file-like that sends every line to a handler.
    """

    def __init__(self, handler, original):
        self.handler = handler
        self.original = original

    def write(self, line):
        """
        Overrides the default IO write(). Writes to console + queue.
        :param line: String printed to stdout, probably with print()
        """
        self.original.write(line)
        self.handler(line)

    def flush(self):
        self.original.flush()

    def isatty(self):
        return False

    def fileno(self):
        return self.original.fileno()


class BaseStdWrapper(object):

    def __init__(
        self, stdout=False, stdout_handler=None, stderr=False, stderr_handler=None
    ):
        self.stdout = stdout
        self.stdout_handler = stdout_handler
        self.stderr = stderr
        self.stderr_handler = stderr_handler

        self._stdout = None
        self._stderr = None

        self._old_stdout = None
        self._old_stderr = None

    def __enter__(self):
        if self.stdout and self.stdout_handler:
            self._stdout = FakeStd(self.stdout_handler, sys.stdout)
            self._old_stdout = sys.stdout
            sys.stdout = self._stdout

        if self.stderr and self.stderr_handler:
            self._stderr = FakeStd(self.stderr_handler, sys.stderr)
            self._old_stderr = sys.stderr
            sys.stderr = self._stderr

    def __exit__(self, exception_type, exception_value, traceback):
        if self.stdout and self._old_stdout:
            sys.stdout = self._old_stdout
            self._old_stdout = None
            self._stdout = None

        if self.stderr and self._old_stderr:
            sys.stderr = self._old_stderr
            self._old_stderr = None
            self._stderr = None


try:
    # Try using a Wurlitzer based wrapper for Mac and Linux
    from wurlitzer import Wurlitzer

    class WurlitzerStdWrapper(Wurlitzer):
        """ A modified Wurltizer class that forward to the original FD and can
        call callbacks for captured streaming data.
        """

        def __init__(self, *args, **kwargs):
            self.stdout_handler = kwargs.pop("stdout_handler", None)
            self.stderr_handler = kwargs.pop("stderr_handler", None)
            super(WurlitzerStdWrapper, self).__init__(*args, **kwargs)
            self.finished = False

        def _handle_stdout(self, data):
            if self._stdout:
                os.write(self._save_fds["stdout"], data)

                if self.stdout_handler:
                    try:
                        self.stdout_handler(self._decode(data))
                    except Exception:
                        # Avoid raising exceptions
                        pass

        def _handle_stderr(self, data):
            if self._stderr:
                os.write(self._save_fds["stderr"], data)

                if self.stderr_handler:
                    try:
                        self.stderr_handler(self._decode(data))
                    except Exception:
                        # Avoid raising exceptions
                        pass

        def _finish_handle(self):
            self.finished = True

    HAS_WURLITZER = True

except ImportError:
    HAS_WURLITZER = False


class StdLogger(object):

    def __init__(self, streamer, wrapper_class):
        self.streamer = streamer
        self.experiment = None
        self.wrapper = wrapper_class(
            stdout=True,
            stdout_handler=self.stdout_handler,
            stderr=True,
            stderr_handler=self.stderr_handler,
        )
        self.wrapper.__enter__()
        self.wrapped = True

    def clean(self):
        if self.wrapped is True:
            # Restore sys.std*
            self.wrapper.__exit__(None, None, None)
            self.wrapped = False

    def set_experiment(self, experiment):
        self.experiment = experiment

    def stdout_handler(self, data):
        self.handler(data, "stdout")

    def stderr_handler(self, data):
        self.handler(data, "stderr")

    def handler(self, data, std_name):
        if not self.experiment:
            return

        payload = self.experiment._create_message()

        if std_name == "stdout":
            payload.set_stdout(data)
        elif std_name == "stderr":
            payload.set_stderr(data)
        else:
            raise NotImplementedError()

        self.streamer.put_messge_in_q(payload)


def get_std_logger(wrapper_spec, streamer):
    """ A factory that pass the right wrapper class to the StdLogger
    """
    if wrapper_spec is None:
        return None
    elif wrapper_spec == "simple":
        return StdLogger(streamer, BaseStdWrapper)
    elif wrapper_spec == "native":
        if HAS_WURLITZER is True:
            return StdLogger(streamer, WurlitzerStdWrapper)
        else:
            LOGGER.warning(NATIVE_STD_WRAPPER_NOT_AVAILABLE)
    else:
        LOGGER.warning(UNKOWN_STD_WRAPPER_SPEC, wrapper_spec)
    return StdLogger(streamer, BaseStdWrapper)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

""" This module handles syncing git repos with the backend. Used for pull
request features."""

import json
import logging
import os
import shutil
import tempfile
import zipfile
from multiprocessing import Process
from os.path import join, splitext

import requests

from ._reporting import FILE_UPLOADED_FAILED, REPO_UPLOADED, REPO_UPLOADED_FAILED
from .config import CONFIG_PATH, read_config_file
from .connection import Reporting, get_backend_session

LOGGER = logging.getLogger(__name__)


def get_repo_root(endpoint, project_id, experiment_id, file_path):
    """
    Gets the git repo path from server.

    Args:
        endpoint: path to server endpoint
        project_id: unique project identifier (required)
        experiment_id: unique experiment identifier (required)
        file_path: current file path. Could be any file that belongs to the
        repo.

    Returns: path to git repo

    """
    payload = {
        "projectId": project_id, "filePath": file_path, "experimentId": experiment_id
    }
    r = get_backend_session().get(endpoint, params=payload)
    ret_val = json.loads(r.text)

    if "root_path" in ret_val and ret_val["root_path"] is not None:
        return ret_val["root_path"]

    elif "msg" in ret_val:
        raise ValueError(ret_val["msg"])

    return None


def merge_config_file(base_config, repo_root_path):
    """ Try reading the configuration file at the root of the repository and
    return the merged config
    Args:
        base_config: a configuration dictionary, usually a copy of the
        default configuration
        repo_root_path: path of repository root
    """
    config_file_path = join(repo_root_path, CONFIG_PATH)

    return read_config_file(base_config, config_file_path)


def compress_py_files(repo_root_path, extensions):
    """
    Compresses all files ending with given extensions in repo to a single zip
    file
    Args:
        repo_root_path: path of folder to zip
        extensions: list of strings containing extensions of files to zip

    Returns: (path to folder that contains zip file, full path to zip file)

    """
    zip_dir = tempfile.mkdtemp()
    zip_path = join(zip_dir, "repo.zip")

    archive = zipfile.ZipFile(zip_path, "w")

    for root, _, files in os.walk(repo_root_path):
        for afile in files:
            extension = splitext(afile)[-1].lower()
            if extension in extensions:
                arcname = join(root.replace(repo_root_path, ""), afile)
                archive.write(join(root, afile), arcname=arcname)
    archive.close()

    return zip_dir, zip_path


def _send_file(url, files, params):
    r = get_backend_session().post(url, params=params, files=files)

    if r.status_code != 200:
        raise ValueError("POSTing file failed: %s" % r.content)


def send_file(
    post_endpoint, api_key, experiment_id, project_id, file_path, additional_params=None
):
    params = {"experimentId": experiment_id, "projectId": project_id, "apiKey": api_key}

    if additional_params is not None:
        params.update(additional_params)

    with open(file_path, "rb") as _file:

        files = {"file": _file}

        _send_file(post_endpoint, params=params, files=files)


def send_file_contents(
    post_endpoint,
    api_key,
    experiment_id,
    project_id,
    file_contents,
    file_name,
    additional_params=None,
):
    params = {"experimentId": experiment_id, "projectId": project_id, "apiKey": api_key}

    if additional_params is not None:
        params.update(additional_params)

    files = {"file": (file_name, file_contents)}

    _send_file(post_endpoint, params=params, files=files)


def upload_repo(
    project_id,
    experiment_id,
    file_path,
    get_path_endpoint,
    post_zip_endpoint,
    api_key,
    config,
):
    """
    Determines repo path and uploads a subset of it's files to server. Used to
    create pull requests on the frontend.
    Args:
        project_id: unique project id
        experiment_id: unique experiment id
        file_path: path to a python file that is part of the repo
        get_path_endpoint: server endpoint url to get repo path
        post_zip_endpoint: server endpoint url to send zip file
        config: a configuration dictionary, usually a copy of the default
        configuration
        api_key: the user's api_key

    Returns: None

    """
    try:
        repo_root_path = get_repo_root(
            get_path_endpoint, project_id, experiment_id, file_path
        )

        if not repo_root_path:
            return

        config = merge_config_file(config, repo_root_path)

        zip_folder, zip_path = compress_py_files(
            repo_root_path, config["uploaded_extensions"]
        )

        send_file(post_zip_endpoint, api_key, experiment_id, project_id, zip_path)
        Reporting.report(
            event_name=REPO_UPLOADED,
            experiment_key=experiment_id,
            project_id=project_id,
            api_key=api_key,
        )

        # Cleanup temp directory
        shutil.rmtree(zip_folder)
    except ValueError as e:
        LOGGER.error("Repo files would not be synced", exc_info=True)
        Reporting.report(
            event_name=REPO_UPLOADED_FAILED,
            experiment_key=experiment_id,
            project_id=project_id,
            api_key=api_key,
            err_msg=str(e),
        )


def upload_file(
    project_id,
    experiment_id,
    file_path,
    upload_endpoint,
    api_key,
    additional_params=None,
    clean=True,
):
    try:
        send_file(
            upload_endpoint,
            api_key,
            experiment_id,
            project_id,
            file_path,
            additional_params,
        )

        if clean is True:
            # Cleanup file
            try:
                os.remove(file_path)
            except OSError:
                pass
    except Exception as e:
        LOGGER.error("File could not be uploaded", exc_info=True)
        Reporting.report(
            event_name=FILE_UPLOADED_FAILED,
            experiment_key=experiment_id,
            project_id=project_id,
            api_key=api_key,
            err_msg=str(e),
        )


def upload_file_contents(
    project_id,
    experiment_id,
    file_contents,
    file_name,
    upload_endpoint,
    api_key,
    additional_params=None,
):
    try:
        send_file_contents(
            upload_endpoint,
            api_key,
            experiment_id,
            project_id,
            file_contents,
            file_name,
            additional_params,
        )
    except Exception as e:
        LOGGER.error("File could not be uploaded", exc_info=True)
        Reporting.report(
            event_name=FILE_UPLOADED_FAILED,
            experiment_key=experiment_id,
            project_id=project_id,
            api_key=api_key,
            err_msg=str(e),
        )


def upload_repo_start_process(
    project_id,
    experiment_id,
    file_path,
    get_path_endpoint,
    post_zip_endpoint,
    api_key,
    config,
):
    args = (
        project_id,
        experiment_id,
        file_path,
        get_path_endpoint,
        post_zip_endpoint,
        api_key,
        config,
    )
    p = Process(target=upload_repo, args=args)
    p.start()
    return p


def upload_file_process(*args):
    p = Process(target=upload_file, args=args)
    p.start()
    return p


def upload_file_contents_process(*args):
    p = Process(target=upload_file_contents, args=args)
    p.start()
    return p



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

"""
Author: Boris Feld

This module contains logging configuration for Comet

"""

import logging
from copy import copy

MSG_FORMAT = "COMET %(levelname)s: %(message)s"
FILE_MSG_FORMAT = "[%(process)d-%(processName)s:%(thread)d] %(relativeCreated)d COMET %(levelname)s [%(filename)s:%(lineno)d]: %(message)s"

GO_TO_DOCS_MSG = " \nFor more details, please refer to: https://www.comet.ml/docs/python-sdk/warnings-errors"

INTERNET_CONNECTION_ERROR = (
    "Failed to establish connection to Comet server. Please check your internet connection. "
    "Your experiment would not be logged"
    + GO_TO_DOCS_MSG
)


INVALID_API_KEY = (
    "The given api key %s is invalid, please check it against the dashboard. "
    "Your experiment would not be logged"
    + GO_TO_DOCS_MSG
)

IPYTHON_NOTEBOOK_WARNING = (
    "Comet.ml support for Ipython Notebook is limited at the moment,"
    " automatic monitoring and stdout capturing is deactivated"
    + GO_TO_DOCS_MSG
)


METRIC_ARRAY_WARNING = (
    "Cannot safely convert %r object to a scalar value, using it string"
    " representation for logging."
)

EXPERIMENT_OPTIMIZER_API_KEY_MISMTACH_WARNING = (
    "WARNING: Optimizer and Experiments API keys mismatch. Please use"
    " the same API key for both."
)


PARSING_ERR_MSG = """We failed to parse your parameter configuration file.

Type casting will be disabled for this run, please fix your configuration file.
"""

CASTING_ERROR_MESSAGE = """Couldn't cast parameter %r, returning raw value instead.
Please report it to comet.ml and use `.raw(%r)` instead of `[%r]` in the meantime."""

NOTEBOOK_MISSING_ID = (
    "We detected that you are running inside a Ipython/Jupyter notebook environment but we cannot save your notebook source code."
    " Please be sure to have installed comet_ml as a notebook server extension by running:\n"
    "jupyter comet_ml enable"
    + GO_TO_DOCS_MSG
)

LOG_IMAGE_OS_ERROR = (
    "We failed to read file %s for uploading.\n"
    "Please double check the file path and permissions"
)

LOG_IMAGE_TOO_BIG = ("File %s is bigger than the upload limit, %s > %s")

LOG_FIGURE_TOO_BIG = ("Figure is bigger than the upload limit, %s > %s")

NATIVE_STD_WRAPPER_NOT_AVAILABLE = (
    "Native output logging mode is not available, fallbacking on basic output logging"
)

UNKOWN_STD_WRAPPER_SPEC = (
    "Unknown output logging mode: %s, fallbacking on basic output logging"
)

EXPERIMENT_LIVE = ("Experiment is live on comet.ml %s\n")


class TracebackLessFormatter(logging.Formatter):

    def format(self, record):
        # Make a copy of the record to avoid altering it
        new_record = copy(record)

        # And delete exception informations so no traceback could be formatted
        # and displayed
        new_record.exc_info = None
        new_record.exc_text = None

        return super(TracebackLessFormatter, self).format(new_record)


WS_ON_OPEN_MSG = "WS Socket connection open"

WS_ON_CLOSE_MSG = "WS connection closed"


def setup(log_file_path=None, file_level=None):
    root = logging.getLogger("comet_ml")
    root.setLevel(logging.DEBUG)

    # Don't send comet-ml to the application logger
    root.propagate = False

    # Add handler
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    console.setFormatter(TracebackLessFormatter(MSG_FORMAT))
    root.addHandler(console)

    # The std* logger might conflicts with the logging if a log record is
    # emitted for each WS message as it would results in an infinite loop. To
    # avoid this issue, all log records after the creation of a message should
    # be at a level lower than info as the console handler is set to info
    # level.

    # Add an additional file handler
    if log_file_path is not None:
        file_handler = logging.FileHandler(log_file_path)

        if file_level is not None:
            file_level = file_level.upper()
            file_handler.setLevel(file_level)
        else:
            file_handler.setLevel(logging.DEBUG)

        file_handler.setFormatter(logging.Formatter(FILE_MSG_FORMAT))
        root.addHandler(file_handler)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import logging
import sys

from comet_ml import config

LOGGER = logging.getLogger(__name__)

SKLEARN_ALREADY_IMPORTED_MSG = (
    "Please import comet before importing any sklearn modules"
)


def fit_logger(real_fit):

    def wrapper(*args, **kwargs):
        ret_val = real_fit(*args, **kwargs)
        try:
            if (
                config.experiment
                and config.experiment.disabled_monkey_patching is False
            ):
                config.experiment.log_multiple_params(ret_val.get_params())
        except Exception:
            LOGGER.error("Failed to extract parameters from estimator")
        finally:
            return ret_val

    return wrapper


def pipeline_fit_logger(real_fit):

    def wrapper(*args, **kwargs):
        ret_val = real_fit(*args, **kwargs)
        try:
            params = ret_val.get_params()
            if params is not None and "steps" in params:
                disabled = config.experiment.disabled_monkey_patching is False
                if config.experiment and disabled:
                    experiment = config.experiment
                    for step in params["steps"]:
                        step_name, step_mdl = step
                        experiment.log_multiple_params(
                            step_mdl.get_params(), prefix=step_name
                        )
        except Exception:
            LOGGER.error("Failed to extract parameters from Pipeline")
        finally:
            return ret_val

    return wrapper


FIT_MODULES = [
    ("sklearn.linear_model.theil_sen", "TheilSenRegressor.fit"),
    ("sklearn.svm.classes", "SVR.fit"),
    ("sklearn.linear_model.stochastic_gradient", "SGDRegressor.fit"),
    ("sklearn.linear_model.ridge", "RidgeCV.fit"),
    ("sklearn.linear_model.ridge", "Ridge.fit"),
    ("sklearn.ensemble.forest", "RandomForestRegressor.fit"),
    ("sklearn.neighbors.regression", "RadiusNeighborsRegressor.fit"),
    ("sklearn.linear_model.ransac", "RANSACRegressor.fit"),
    ("sklearn.linear_model.passive_aggressive", "PassiveAggressiveRegressor.fit"),
    ("sklearn.cross_decomposition.pls_", "PLSRegression.fit"),
    ("sklearn.cross_decomposition.pls_", "PLSCanonical.fit"),
    ("sklearn.linear_model.omp", "OrthogonalMatchingPursuitCV.fit"),
    ("sklearn.linear_model.omp", "OrthogonalMatchingPursuit.fit"),
    ("sklearn.svm.classes", "NuSVR.fit"),
    ("sklearn.linear_model.coordinate_descent", "MultiTaskLassoCV.fit"),
    ("sklearn.linear_model.coordinate_descent", "MultiTaskLasso.fit"),
    ("sklearn.linear_model.coordinate_descent", "MultiTaskElasticNetCV.fit"),
    ("sklearn.linear_model.coordinate_descent", "MultiTaskElasticNet.fit"),
    ("sklearn.neural_network.multilayer_perceptron", "MLPRegressor.fit"),
    ("sklearn.svm.classes", "LinearSVR.fit"),
    ("sklearn.linear_model.base", "LinearRegression.fit"),
    ("sklearn.linear_model.least_angle", "LassoLarsIC.fit"),
    ("sklearn.linear_model.least_angle", "LassoLarsCV.fit"),
    ("sklearn.linear_model.least_angle", "LassoLars.fit"),
    ("sklearn.linear_model.coordinate_descent", "LassoCV.fit"),
    ("sklearn.linear_model.coordinate_descent", "Lasso.fit"),
    ("sklearn.linear_model.least_angle", "LarsCV.fit"),
    ("sklearn.linear_model.least_angle", "Lars.fit"),
    ("sklearn.kernel_ridge", "KernelRidge.fit"),
    ("sklearn.neighbors.regression", "KNeighborsRegressor.fit"),
    ("sklearn.linear_model.huber", "HuberRegressor.fit"),
    ("sklearn.ensemble.gradient_boosting", "GradientBoostingRegressor.fit"),
    ("sklearn.gaussian_process.gpr", "GaussianProcessRegressor.fit"),
    ("sklearn.gaussian_process.gaussian_process", "GaussianProcess.fit"),
    ("sklearn.ensemble.forest", "ExtraTreesRegressor.fit"),
    ("sklearn.tree.tree", "ExtraTreeRegressor.fit"),
    ("sklearn.linear_model.coordinate_descent", "ElasticNetCV.fit"),
    ("sklearn.linear_model.coordinate_descent", "ElasticNet.fit"),
    ("sklearn.tree.tree", "DecisionTreeRegressor.fit"),
    ("sklearn.cross_decomposition.cca_", "CCA.fit"),
    ("sklearn.linear_model.bayes", "BayesianRidge.fit"),
    ("sklearn.ensemble.bagging", "BaggingRegressor.fit"),
    ("sklearn.ensemble.weight_boosting", "AdaBoostRegressor.fit"),
    ("sklearn.linear_model.bayes", "ARDRegression.fit"),
    ("sklearn.svm.classes", "SVC.fit"),
    ("sklearn.linear_model.ridge", "RidgeClassifierCV.fit"),
    ("sklearn.linear_model.ridge", "RidgeClassifier.fit"),
    ("sklearn.ensemble.forest", "RandomForestClassifier.fit"),
    ("sklearn.neighbors.classification", "RadiusNeighborsClassifier.fit"),
    ("sklearn.discriminant_analysis", "QuadraticDiscriminantAnalysis.fit"),
    ("sklearn.linear_model.perceptron", "Perceptron.fit"),
    ("sklearn.linear_model.passive_aggressive", "PassiveAggressiveClassifier.fit"),
    ("sklearn.svm.classes", "NuSVC.fit"),
    ("sklearn.neighbors.nearest_centroid", "NearestCentroid.fit"),
    ("sklearn.naive_bayes", "MultinomialNB.fit"),
    ("sklearn.neural_network.multilayer_perceptron", "MLPClassifier.fit"),
    ("sklearn.linear_model.logistic", "LogisticRegressionCV.fit"),
    ("sklearn.linear_model.logistic", "LogisticRegression.fit"),
    ("sklearn.svm.classes", "LinearSVC.fit"),
    ("sklearn.discriminant_analysis", "LinearDiscriminantAnalysis.fit"),
    ("sklearn.semi_supervised.label_propagation", "LabelSpreading.fit"),
    ("sklearn.semi_supervised.label_propagation", "LabelPropagation.fit"),
    ("sklearn.neighbors.classification", "KNeighborsClassifier.fit"),
    ("sklearn.ensemble.gradient_boosting", "GradientBoostingClassifier.fit"),
    ("sklearn.gaussian_process.gpc", "GaussianProcessClassifier.fit"),
    ("sklearn.naive_bayes", "GaussianNB.fit"),
    ("sklearn.ensemble.forest", "ExtraTreesClassifier.fit"),
    ("sklearn.tree.tree", "ExtraTreeClassifier.fit"),
    ("sklearn.tree.tree", "DecisionTreeClassifier.fit"),
    ("sklearn.calibration", "CalibratedClassifierCV.fit"),
    ("sklearn.naive_bayes", "BernoulliNB.fit"),
    ("sklearn.ensemble.bagging", "BaggingClassifier.fit"),
    ("sklearn.ensemble.weight_boosting", "AdaBoostClassifier.fit"),
    ("sklearn.linear_model.stochastic_gradient", "SGDClassifier.fit"),
]

PIPELINE_FIT_MODULES = [("sklearn.pipeline", "Pipeline.fit")]


def patch(module_finder):
    if "sklearn" in sys.modules:
        LOGGER.warning(SKLEARN_ALREADY_IMPORTED_MSG)

    # Register the pipeline fit methods
    for module, object_name in PIPELINE_FIT_MODULES:
        module_finder.register(module, object_name, pipeline_fit_logger)

    # Register the fit methods
    for module, object_name in FIT_MODULES:
        module_finder.register(module, object_name, fit_logger)


# https://blog.sqreen.io/dynamic-instrumentation-agent-for-python/



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

""" The import hook which monkey patch modules
"""

import imp
import sys

from ._notebook import in_notebook_environment


def get_pep_302_importer(path):
    """ A pep 302 compatible importer
    """
    importer = sys.path_importer_cache.get(path)
    if not importer:
        for hook in sys.path_hooks:
            try:
                importer = hook(path)
                break
            except ImportError:
                pass
    return importer


class CustomFileLoader(object):
    """ A Python 3 loader that use a SourceFileLoader to exec an imported
    module and patch it with CometModuleFinder
    """

    def __init__(self, loader, fullname, finder):
        self.loader = loader
        self.fullname = fullname
        self.finder = finder

    def exec_module(self, module):
        # Execute the module source code to define all the objects
        if hasattr(self.loader, "exec_module"):
            self.loader.exec_module(module)
        else:
            # zipimporter doesn't use exec_module
            module = self.loader.load_module(self.fullname)

        return self.finder._patch(module, self.fullname)

    def create_module(self, spec):
        """ Mandatory in Python 3.6 as we define the exec_module method
        """
        return None


class CometModuleFinder(object):

    def __init__(self):
        self.patcher_functions = {}

        if sys.version_info[0] >= 3:
            from importlib.machinery import PathFinder

            self.pathfinder = PathFinder()

    def register(self, module_name, object_name, patcher_function):
        module_patchers = self.patcher_functions.setdefault(module_name, {})
        module_patchers[object_name] = patcher_function

    def start(self):
        if self not in sys.meta_path and not in_notebook_environment():
            sys.meta_path.insert(0, self)

    def find_module(self, fullname, path=None):
        """ Python 2 import hook
        """
        if fullname not in self.patcher_functions:
            return

        return self

    def load_module(self, fullname):
        """ Python 2 import hook
        """
        module = self._get_module(fullname)
        return self._patch(module, fullname)

    def find_spec(self, fullname, path=None, target=None):
        """ Python 3 import hook
        """
        if fullname not in self.patcher_functions:
            return

        from importlib.machinery import ModuleSpec

        spec = self.pathfinder.find_spec(fullname, path, target)

        if not spec:
            return

        return ModuleSpec(fullname, CustomFileLoader(spec.loader, fullname, self))

    def _get_module(self, fullname):
        splitted_name = fullname.split(".")
        parent = ".".join(splitted_name[:-1])

        if fullname in sys.modules:
            return sys.modules[fullname]
        elif parent in sys.modules:
            parent = sys.modules[parent]
            module_path = imp.find_module(splitted_name[-1], parent.__path__)
            return imp.load_module(fullname, *module_path)
        else:
            try:
                module_path = imp.find_module(fullname)
                return imp.load_module(fullname, *module_path)
            except ImportError:
                for p in sys.path:
                    importer = get_pep_302_importer(p)

                    # Ignore invalid paths
                    if importer is None:
                        continue

                    module_path = importer.find_module(fullname)

                    if module_path:
                        return importer.load_module(fullname)

    def _patch(self, module, fullname):
        objects_to_patch = self.patcher_functions.get(fullname, {})

        for object_name, patcher_function in objects_to_patch.items():
            object_path = object_name.split(".")

            original = self._get_object(module, object_path)

            if original is None:
                # TODO: Send back the error?
                continue

            new_object = patcher_function(original)
            self._set_object(module, object_path, new_object)

        return module

    def _get_object(self, module, object_path):
        current_object = module

        for part in object_path:
            try:
                current_object = getattr(current_object, part)
            except AttributeError:
                return None

        return current_object

    def _set_object(self, module, object_path, new_object):
        object_to_patch = self._get_object(module, object_path[:-1])
        setattr(object_to_patch, object_path[-1], new_object)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

"""
Author: Gideon Mendels

This module contains the functions related to Ipython/Jupyter notebook support
"""

import copy
import errno
import hashlib
import json
import logging
import os
from base64 import b64encode
from os.path import abspath, expanduser, isdir, isfile, join

from IPython import get_ipython
from notebook.base.handlers import IPythonHandler
from notebook.utils import url_path_join

from .connection import notebook_source_upload

LOGGER = logging.getLogger(__name__)
CACHE_DIR = expanduser("~/.comet_notebook_cache/")
NOTEBOOK_MAPPING = {}


def _jupyter_server_extension_paths():
    return [{"module": "comet_ml"}]


def _jupyter_nbextension_paths():
    return []


def load_jupyter_server_extension(nbapp):
    nbapp.log.info("Comet.ml extension activated!")

    contents_manager = nbapp.contents_manager
    contents_manager.pre_save_hook = pre_save_hook

    web_app = nbapp.web_app

    # make sure our static files are available
    static_files_path = os.path.normpath(
        os.path.join(os.path.dirname(__file__), "static")
    )
    LOGGER.debug("Editing nbextensions path to add %s", static_files_path)
    if static_files_path not in web_app.settings["nbextensions_path"]:
        web_app.settings["nbextensions_path"].append(static_files_path)

    host_pattern = ".*$"
    web_app.add_handlers(
        host_pattern,
        [
            (
                url_path_join(web_app.settings["base_url"], r"/comet_ml"),
                NotebookHashMapperHandler,
            )
        ],
    )


def get_notebook_id():
    _Jupyter = get_ipython()
    if _Jupyter is None:
        return None

    try:
        return _Jupyter.kernel.shell.user_ns["NOTEBOOK_ID"]

    except KeyError:
        return None


def parse_jupyter_server_model(model):
    api_key = None

    for cell in model["content"]["cells"]:
        if cell["cell_type"] != "code":
            continue

        # Clean up the execution count in order to have stable hashes
        cell["execution_count"] = None

        for output in cell["outputs"]:
            output["execution_count"] = None

        for line in cell["source"].splitlines():
            if line.startswith("COMET_ML_API_KEY="):
                api_key = line[len("COMET_ML_API_KEY=") + 1:-1]

    return model, api_key


def get_hash_content(model):
    json_model = json.dumps(model, sort_keys=True)
    content_hash = hashlib.sha1(json_model.encode("utf-8")).hexdigest()

    return content_hash, json_model


def create_cache_dir():
    try:
        os.makedirs(CACHE_DIR)
    except (IOError, OSError) as error:
        if error.errno == errno.EEXIST and isdir(CACHE_DIR):
            return

        raise


def pre_save_hook(model, **kwargs):
    local_path = kwargs.pop("path", None)

    if local_path is None:
        LOGGER.warning("No notebook path given, cannot save it")
        return

    path = abspath(local_path)

    if model["type"] != "notebook":
        return

    # only run on nbformat v4
    if model["content"]["nbformat"] != 4:
        return

    new_model, api_key = parse_jupyter_server_model(copy.deepcopy(model))
    if api_key is None:
        LOGGER.info("Couldn't detect an api key in the notebook")
        return

    content_hash, json_model = get_hash_content(new_model)

    NOTEBOOK_MAPPING[path] = content_hash

    cache_file_name = "%s_%s" % (b64encode(path.encode("utf-8")), content_hash)
    cache_file_path = join(CACHE_DIR, cache_file_name)
    if isfile(cache_file_path):
        return

    notebook_source_upload(content_hash, json_model, api_key, path)

    try:
        create_cache_dir()

        # Create the empty cache file
        with open(cache_file_path, "a"):
            pass
    except IOError:
        LOGGER.warning("Failed to create a cache file", exc_info=True)


class NotebookHashMapperHandler(IPythonHandler):

    def get(self):
        local_path = self.get_argument("jupyter_path", "")
        path = abspath(local_path)
        msg_json = json.dumps({"NOTEBOOK_HASH_ID": NOTEBOOK_MAPPING.get(path)})
        self.write(msg_json)
        self.flush()



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import logging
import sys

from comet_ml import config, tensorboard_logger

LOGGER = logging.getLogger(__name__)


def fit_logger(real_fit):

    def wrapper(*args, **kwargs):
        if config.experiment and config.experiment.disabled_monkey_patching is False:
            callback = config.experiment.get_keras_callback()
            if "callbacks" in kwargs and kwargs["callbacks"] is not None:
                callbacks = kwargs["callbacks"]
                # Only append the callback if it's not there.
                if not any(x.__class__ == callback.__class__ for x in callbacks):
                    callbacks.append(callback)
            else:
                kwargs["callbacks"] = [callback]

            # Disable tensorboard logging as it conflicts with keras
            tensorboard_logger.DISABLED = True

        return real_fit(*args, **kwargs)

    return wrapper


def multi_gpu_model_wrapper(real_multi_gpu_model):

    def wrapper(model, *args, **kwargs):
        result = real_multi_gpu_model(model, *args, **kwargs)

        # multi_gpu_model should always return a new Model
        try:
            result._smodel_json = model.to_json()
        except Exception:
            LOGGER.info("Failed to saved multi-GPU model")
            result._smodel_json = None

        return result

    return wrapper


def patch(module_finder):
    module_finder.register("keras.models", "Model.fit", fit_logger)
    module_finder.register("keras.models", "Model.fit_generator", fit_logger)
    module_finder.register("tensorflow.python.keras.models", "Model.fit", fit_logger)
    module_finder.register(
        "tensorflow.python.keras.models", "Model.fit_generator", fit_logger
    )
    module_finder.register(
        "keras.utils.training_utils", "multi_gpu_model", multi_gpu_model_wrapper
    )


if "keras" in sys.modules:
    raise SyntaxError("Please import Comet before importing any keras modules")



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************
#
"""
Author: Boris Feld

This module contains the various helpers for the Optimization API
"""
import logging

from comet_ml import config

from ._logging import CASTING_ERROR_MESSAGE
from .exceptions import OptimizationMissingExperiment, PCSCastingError, PCSParsingError

LOGGER = logging.getLogger(__name__)


class Suggestion(object):
    """ A suggestion is a single proposition of hyper-parameters values.

    You can use it like a dict:

    ```python
    suggestion["x"] # Returns the value for hyper-parameter x
    ```

    Suggestion is automatically casting values for hyper-parameter declared as
    `integer` or `real`. It will returns `int` for `integer` and `float` for
    `real`. For `categorical` and `ordinal`, it will returns `str`.

    In case casting is failing, it will print a warning message and return a `str`.

    For accessing the raw value without casting, you can use the `raw` method:

    ```python
    suggestion.raw("x") # Always returns a str
    ```
    """

    def __init__(self, suggestion, optimizer, types):
        """ You shouldn't directly instantiate Suggestion objects, use
        [Optimizer.get_suggestion](/Optimizer/#optimizerget_suggestion)
        instead.
        """
        self.suggestion = suggestion
        self.run_id = suggestion["run_id"]
        self.params = suggestion["params"]
        self.optimizer = optimizer
        self.types = types

    def __iter__(self):
        return iter(self.params)

    def __getitem__(self, name):
        """ Return the casted value for this hyper-parameter.
        Args:
            name: The hyper-parameter name
        """
        raw_value = self.params[name]
        try:
            return cast_parameter(raw_value, self.types[name])
        except (KeyError, PCSCastingError):
            LOGGER.warning(CASTING_ERROR_MESSAGE, name, name, name)
            return self.raw(name)

    def raw(self, name):
        """ Return the raw not-casted value for this hyper-parameter.
        Args:
            name: The hyper-parameter name
        """
        return self.params[name]

    def report_score(self, name, score):
        """ Send back the score for this suggestion.
        Args:
            score: A float representing the score
        """
        self._report_params_to_experiment(self.params, name, score)

        self.optimizer._report_score(self.run_id, score)

    def _report_params_to_experiment(self, suggestion, name, score):
        if config.experiment is None:
            raise OptimizationMissingExperiment

        exp = config.experiment

        exp.log_multiple_params(suggestion)
        exp.log_metric(name, score)


PCS_TYPES = {"integer", "real", "ordinal", "categorical"}


def parse_pcs(pcs_content):
    parsed = {}
    for line in pcs_content.splitlines():
        # Clean line
        line = line.strip()

        # Ignore empty lines
        if line == "":
            continue

        # Ignore commented lines
        if line.startswith("#"):
            continue

        # Ignore conditions as they doesn't influence parameters type
        if "|" in line:
            continue

        # Ignore forbidden parameter syntax
        if line.startswith("{") and line.endswith("}"):
            continue

        # Check that the line looks valid
        if "}" not in line and "]" not in line:
            raise PCSParsingError(line)

        splitted = line.split(" ")

        if len(splitted) < 2:
            raise PCSParsingError(line)

        param_name = splitted[0]
        param_type = splitted[1]

        # Check that type is valid
        if param_type not in PCS_TYPES:
            raise PCSParsingError(line)

        parsed[param_name] = param_type

    return parsed


def cast_parameter(value, pcs_type):
    if pcs_type not in PCS_TYPES:
        raise PCSCastingError(value, pcs_type)

    if pcs_type == "integer":
        try:
            return int(value)
        except ValueError:
            raise PCSCastingError(value, pcs_type)
    elif pcs_type == "real":
        try:
            return float(value)
        except ValueError:
            raise PCSCastingError(value, pcs_type)
    elif pcs_type == "categorical":
        try:
            return int(value)
        except ValueError:
            pass

        try:
            return float(value)
        except ValueError:
            pass

        try:
            return _str_to_bool(value)
        except ValueError:
            pass

    return value


def _str_to_bool(s):
    s = s.lower()
    if s == "true":
        return True
    elif s == "false":
        return False
    else:
        raise ValueError()



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

"""
Extends keras Callbacks. Provides automatic logging and tracking with Comet.ml
"""

import logging

try:
    from keras import callbacks
except ImportError:
    from tensorflow.contrib.keras import callbacks


LOGGER = logging.getLogger(__name__)


class EmptyKerasCallback(callbacks.Callback):
    """
    Empty Keras callback. TODO(gidim): remove this
    """

    def __init__(self):
        super(EmptyKerasCallback, self).__init__()

    def on_epoch_begin(self, epoch, logs=None):
        pass

    def on_epoch_end(self, epoch, logs=None):
        pass

    def on_batch_begin(self, batch, logs=None):
        pass

    def on_batch_end(self, batch, logs=None):
        pass

    def on_train_begin(self, logs=None):
        pass

    def on_train_end(self, logs=None):
        pass


class KerasCallback(callbacks.Callback):
    """ Keras callback to report params, metrics to Comet.ml Experiemnt()"""

    def __init__(self, experiment, log_params=True, log_metrics=True, log_graph=True):
        """
        Create a new experiment and submit source code.
        :param api_key: User's API key. Required.
        """
        super(KerasCallback, self).__init__()

        # Inits the experiment with reference to the name of this class. Required for loading the correct
        # script file
        self.experiment = experiment
        self.log_params = log_params
        self.log_metrics = log_metrics
        self.log_graph = log_graph
        self.our_step = 0
        self._ignores = ["verbose", "do_validation"]

    def on_epoch_begin(self, epoch, logs=None):
        LOGGER.debug("On epoch begin %s %s", epoch, logs)

    def on_epoch_end(self, epoch, logs=None):
        LOGGER.debug("On epoch end %s %s", epoch, logs)
        if self.log_metrics:
            self.experiment.log_epoch_end(epoch, step=self.our_step)
            if logs:
                for name, val in logs.items():
                    self.experiment.log_metric(name, val, step=self.our_step)

    def on_batch_begin(self, batch, logs=None):
        LOGGER.debug("On batch begin %s %s", batch, logs)

    def on_batch_end(self, batch, logs=None):
        """
        Logs training metrics.
        """
        LOGGER.debug("On batch end %s %s", batch, logs)

        self.our_step += 1

        if logs and self.log_metrics:
            for name, val in logs.items():
                self.experiment.log_metric(name, val, step=self.our_step)

    def on_train_begin(self, logs=None):
        """
        Sets model graph.
        """
        LOGGER.debug("On train begin %s", logs)

        if self.log_graph:
            self.experiment.set_model_graph(get_keras_json_model(self.model))
        self.experiment.log_other("trainable_params", self.model.count_params())

        if self.log_params:
            if logs:
                for k, v in logs.items():
                    self.experiment.log_parameter(k, v)

            # Keras Callback doesn't set this parameter at creation by default
            if hasattr(self, "params") and self.params:
                for k, v in self.params.items():
                    if k != "metrics" and k not in self._ignores:
                        self.experiment.log_parameter(k, v)

    def on_train_end(self, *args, **kwargs):
        LOGGER.debug("On train end %r", locals())


def get_keras_json_model(model):
    # With multi-gpu models we save the original model in the _smodel attribute
    smodel_json = getattr(model, "_smodel_json", None)

    if smodel_json is not None:
        return smodel_json

    else:
        try:
            return model.to_json()

        except Exception:
            LOGGER.warning("Failed to save keras graph model")



import setuptools

setuptools.setup(
    name="comet_ml",
    packages=["comet_ml"],
    version="1.0.17",
    url="https://www.comet.ml",
    author="Comet ML Inc.",
    author_email="mail@comet.ml",
    description="Supercharging Machine Learning",
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    install_requires=[
        "websocket-client>=0.44.0",
        "requests>=2.18.4",
        "six",
        "wurlitzer>=1.0.2",
        "netifaces>=0.10.7",
    ],
    include_package_data=True,
    test_requires=["websocket-server", "pytest", "responses"],
    scripts=["comet_ml/scripts/jupyter-comet_ml"],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
    ],
)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

"""
Author: Gideon Mendels

This module contains the functions related to Ipython/Jupyter notebook support
"""

import copy
import hashlib
import json
import os
import sys

NOTEBOOK_MAPPING = {}

try:
    import notebook
    import IPython
    from notebook.utils import url_path_join
    from notebook.base.handlers import IPythonHandler
    from IPython.core.magics.namespace import NamespaceMagics
    from IPython import get_ipython

    from ._real_notebook import (
        _jupyter_server_extension_paths,
        _jupyter_nbextension_paths,
        load_jupyter_server_extension,
        get_notebook_id,
    )

    HAS_NOTEBOOK = True
except ImportError:
    HAS_NOTEBOOK = False

    from ._not_notebook import (
        _jupyter_server_extension_paths,
        _jupyter_nbextension_paths,
        load_jupyter_server_extension,
        get_notebook_id,
    )

IN_NOTEBOOK_ENV = "ipykernel" in sys.modules


def in_notebook_environment():
    return IN_NOTEBOOK_ENV



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import json

convert_functions = []


try:
    import numpy

    def convert_numpy_array(value):
        try:
            return numpy.asscalar(value)

        except (ValueError, IndexError, AttributeError, TypeError):
            return

    convert_functions.append(convert_numpy_array)
except ImportError:
    pass


class NestedEncoder(json.JSONEncoder):
    """
    A JSON Encoder that converts floats/decimals to strings and allows nested objects
    """

    def default(self, obj):

        # First convert the object
        obj = self.convert(obj)

        # Check if the object is convertible
        try:
            json.JSONEncoder().encode(obj)
            return obj

        except TypeError:
            pass

        # Custom conversion
        if obj.__class__.__name__ == "type":
            return str(obj)

        elif hasattr(obj, "repr_json"):
            return obj.repr_json()

        elif isinstance(obj, complex):
            return str(obj)

        else:
            try:
                return json.JSONEncoder.default(self, obj)

            except TypeError as e:
                return "%s not JSON serializable" % obj.__class__.__name__

    def floattostr(self, o, _inf=float("Inf"), _neginf=-float("-Inf"), nan_str="None"):
        if o != o:
            return nan_str

        else:
            return o.__repr__()

    def convert(self, obj):
        """
        Try converting the obj to something json-encodable
        """
        for converter in convert_functions:
            converted = converter(obj)

            if converted is not None:
                obj = converted

        return obj



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

"""
Author: Gideon Mendels

This module contains the functions needed for Jupyter extensions in case we
are not in Jupyter server extension environment
"""

import copy
import hashlib
import json
import os
import sys


def _jupyter_server_extension_paths():
    return []


def _jupyter_nbextension_paths():
    return []


def load_jupyter_server_extension(nbapp):
    return


def get_notebook_id():
    return None
